"""Plot distance histograms."""

import numpy as np
from rnavigate import plots


class DistHist(plots.Plot):
    """Create a distance histogram plot.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    **kwargs
        Keyword arguments passed to :class:`rnavigate.plots.Plot`.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    axes2 : dict
        Dictionary of twin axes objects. Keys are axes objects from
        :attr:`rnavigate.plots.DistHist.axes`.
    i : int
        Index of current axes object. Increments with each call to
        :meth:`rnavigate.plots.DistHist.plot_data`.
    """

    def __init__(self, num_samples, **plot_kwargs):
        """Initialize DistHist object."""
        super().__init__(num_samples, sharey=True, **plot_kwargs)
        self.axes2 = {}
        base_ax2 = None
        for row in self.axes:
            for ax in row:
                ax2 = ax.twinx()
                self.axes2[ax] = ax2
                if base_ax2 is None:
                    base_ax2 = ax2
                else:
                    ax2.sharey(base_ax2)

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=2,
        height_ax_in=2,
        height_gap_in=1,
        width_gap_in=0.4,
        top_in=1,
        bottom_in=1,
        left_in=1,
        right_in=1,
    ):
        """Set figure size.

        Parameters
        ----------
        height_ax_rel : float, defaults to None
            Height of axes objects relative to y-axis limits.
        width_ax_rel : float, defaults to None
            Width of axes objects relative to x-axis limits.
        width_ax_in : float, defaults to 2
            Width of axes objects in inches.
        height_ax_in : float, defaults to 2
            Height of axes objects in inches.
        height_gap_in : float, defaults to 1
            Height of gap between axes objects in inches.
        width_gap_in : float, defaults to 0.4
            Width of gap between axes objects in inches.
        top_in : float, defaults to 1
            Height of top margin in inches.
        bottom_in : float, defaults to 1
            Height of bottom margin in inches.
        left_in : float, defaults to 1
            Width of left margin in inches.
        right_in : float, defaults to 1
            Width of right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )

    def plot_data(
        self, structure, interactions, bg_interactions, label, atom="O2'", ax=None
    ):
        """Plot data on the current (or specified) axes object.

        Parameters
        ----------
        structure : rnavigate.data.SecondaryStructure or rnavigate.data.PDB
            Structure object to compute contact distances or 3D distances.
        interactions : rnavigate.data.Interactions
            Filtered Interactions object to to visualize pairwise distances.
        bg_interactions : rnavigate.data.Interactions
            Filtered Interactions object to to visualize pairwise background distances.
        label : str
            Label for the current axes object.
        atom : str, defaults to "O2'"
            Atom to compute distances from.
        ax : matplotlib.axes.Axes, defaults to None
            Axes object to plot on. If None, use the current axes object.
        """
        if ax is None:
            ax = self.get_ax()
        ax2 = self.axes2[ax]
        if bg_interactions is not None:
            self.plot_experimental_distances(
                ax=ax2,
                structure=structure,
                interactions=bg_interactions,
                atom=atom,
                histtype="step",
            )
        else:
            self.plot_structure_distances(ax=ax2, structure=structure, atom=atom)
        self.plot_experimental_distances(
            ax=ax, structure=structure, interactions=interactions, atom=atom
        )
        ax.set(title=label)
        self.i += 1
        if self.i == self.length:
            for leftmost_axis in self.axes[:, 0]:
                leftmost_axis.set(ylabel="Experimental")
            for rightmost_axis in self.axes[:, -1]:
                self.axes2[rightmost_axis].set(ylabel="Pairwise")
            for top_axis in self.axes[-1, :]:
                top_axis.set(xlabel="3D distance")
            for row in self.axes:
                for not_rightmost_axis in row[:-1]:
                    self.axes2[not_rightmost_axis].yaxis.set_tick_params(
                        labelright=False
                    )

    def plot_structure_distances(self, ax, structure, atom):
        """Plot all distances in the structure.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to plot on.
        structure : rnavigate.data.SecondaryStructure or rnavigate.data.PDB
            Structure object to compute contact distances or 3D distances.
        atom : str
            Atom to compute distances from.
        """
        matrix = structure.get_distance_matrix(atom=atom)
        dists = []
        for i in range(len(matrix) - 6):
            for j in range(i + 6, len(matrix)):
                if not np.isnan(matrix[i, j]):
                    dists.append(matrix.item(i, j))
        ax.hist(
            dists,
            bins=range(0, int(max(dists)) + 5, 5),
            histtype="step",
            color="0.5",
            label="All distances",
        )

    def plot_experimental_distances(
        self, ax, structure, interactions, atom, histtype="bar"
    ):
        """Plot pairwise distances from the interactions object.

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to plot on.
        structure : rnavigate.data.SecondaryStructure or rnavigate.data.PDB
            Structure object to compute contact distances or 3D distances.
        interactions : rnavigate.data.Interactions
            Filtered Interactions object to to visualize pairwise distances.
        atom : str
            Atom to compute distances from.
        histtype : "bar" or "step", defaults to "bar"
            Type of histogram to plot.
        """
        interactions.set_3d_distances(structure, atom)
        ij_dists = interactions.data.loc[interactions.data["mask"], "Distance"]
        if (len(ij_dists) > 0) and (histtype == "bar"):
            ax.hist(
                ij_dists, bins=range(0, int(max(ij_dists)) + 5, 5), width=5, ec="none"
            )
        elif (len(ij_dists) > 0) and (histtype == "step"):
            ax.hist(
                ij_dists,
                bins=range(0, int(max(ij_dists)) + 5, 5),
                histtype="step",
                color="0.5",
                label="All distances",
            )



"""The RNAvigate Sample object for parsing, assigning, and organizing data."""

from rnavigate import data
from rnavigate.data_loading import create_data


class Sample:
    """Loads and organizes RNA structural data for use with plotting functions.

    The Sample class stores all of the relevant experimental and computational
    structural data for a single RNA experiment. Between samples, common data
    types should be given a common data keyword so they can be easily compared.

    Parameters
    ----------
    sample : str
        an arbitrary name. This will be used as a label in plot legends and titles
        to differentiate it from other samples
    inherit : Sample or list of these, optional
        Data keywords and associated data from other samples become data keywords
        and associated data of this sample. This does not make additional copies
        of the data: i.e. operations that make changes to inherited data change the
        original sample, and any other samples that inherited that data. This can
        be useful to save time and memory on operations and large data structures
        that are shared between samples.
    keep_inherited_defaults : bool, default = True
        whether to keep inherited default keywords
    **data_keywords:
        There are many built-in data keywords with different expectations and
        behaviors. For a full list with input formats and output behavior, visit:
        https://rnavigate.readthedocs.io/en/latest/loading-data/

    Attributes
    ----------
    sample : str
        the name of the sample
    inputs : dict
        a dictionary of data keywords and their (user-defined) inputs
    data : dict
        a dictionary of data keywords and their associated data
    defaults : dict
        a dictionary of data classes and their default data keywords

    Example
    -------
    >>> sample = rnavigate.Sample(
    ...     sample="My sample",
    ...     shapemap="path/to/shapmapper_profile.txt",
    ...     ss="path/to/structure.ct",
    ...     ringmap="path/to/ringmapper_rings.txt",
    ...     pdb="path/to/pdb.pdb",
    ...     arbitrary_keyword={
    ...         "sites": [10, 20, 30],
    ...         "name": "sites of interest",
    ...         "color": "red",
    ...     },
    ... )
    >>> sample.print_data_keywords()
    My sample data keywords:
      annotations:
        arbitrary_keyword (default)
      profiles:
        shapemap (default)
      structures:
        ss (default)
      interactions:
        ringmap (default)
      pdbs:
        pdb (default)
    """

    def __init__(
        self, sample, inherit=None, keep_inherited_defaults=True, **data_keywords
    ):
        """Creates a Sample."""
        self.sample = sample
        self.inputs = {}
        self.data = {}
        self.defaults = {
            "default_annotation": None,
            "default_profile": None,
            "default_structure": None,
            "default_interactions": None,
            "default_pdb": None,
        }

        # inherit data from other Samples
        self.inherit_data(
            inherit=inherit,
            keep_inherited_defaults=keep_inherited_defaults,
            overwrite=False,
        )

        # get data and store inputs for all data_keywords
        for data_keyword, inputs in data_keywords.items():
            self.set_data(data_keyword=data_keyword, inputs=inputs)
            self.set_as_default(data_keyword=data_keyword, overwrite=False)

    def inherit_data(self, inherit, keep_inherited_defaults, overwrite):
        """retrieves and stores data and data keywords from other samples

        Parameters:
            inherit : Sample or list of Samples
                Other samples from which to inherit data and data keywords
            keep_inherited_defaults : bool
                Use default values from inherited samples
            overwrite : bool
                whether to overwrite any existing keywords with inherited keywords
        """
        if isinstance(inherit, (list, tuple)):
            for inherit_sample in inherit[::-1]:
                self.inherit_data(inherit_sample, keep_inherited_defaults, overwrite)
        elif isinstance(inherit, Sample):
            if overwrite:
                self.data |= inherit.data
                self.inputs |= inherit.inputs
            else:
                self.data = inherit.data | self.data
                self.inputs = inherit.inputs | self.inputs
            if keep_inherited_defaults and overwrite:
                self.defaults |= inherit.defaults
            elif keep_inherited_defaults:
                for k, v in inherit.defaults.items():
                    if self.defaults[k] is None:
                        self.defaults[k] = v
        elif inherit is not None:
            raise ValueError("inherit only accepts rnav.Sample or list of rnav.Sample")

    def set_data(self, data_keyword, inputs, overwrite=False):
        """Add data to Sample using the given data keyword and inputs

        This methods works similarly to the data keywords arguments used
        during Sample initialization:

            my_sample = rnavigate.Sample(
                sample="name",
                data_keyword=inputs)

        is equivalent to:

            my_sample = rnavigate.Sample(
                sample="name")
            my_sample.add_data(
                "data_keyword", inputs)

        Parameters
        ----------
            data_keyword : str
                a data keyword used to store and/or parse the inputs
            inputs : dict or rnavigate.data.Data
                a dictionary used to create the data object or a data object itself
            overwrite : bool, defaults to False
                whether to overwrite a pre-existing data_keyword
        """
        if (data_keyword in self.data) and not overwrite:
            raise ValueError(
                f"'{data_keyword}' is already a data keyword. "
                "Choose a different one."
            )
        try:
            self.data[data_keyword] = create_data(
                sample=self, data_keyword=data_keyword, inputs=inputs
            )
        except BaseException as e:
            raise ValueError(f"issue while loading {data_keyword}") from e
        self.inputs[data_keyword] = inputs

    def get_data(self, data_keyword, data_class=None):
        """Replaces data keyword with data object, even if nested.

        Parameters
        ----------
            data_keyword : rnavigate.data.Data or data keyword or list/dict of these
                If None, returns None.
                If a data keyword, returns associated data from sample
                If Data, returns that data.
                If a list or dictionary, returns list or dictionary with
                    data keyword values replaced with associated Data
            data_class : rnavigate.data.Data class or subclass, optional
                If provided, ensures that returned data is of this type.

        Returns
        -------
            Same type as data_keyword argument, but data keywords are replaced
                with associated data

        Raises
        ------
            ValueError:
                if data is not found in sample
            ValueError:
                if the data retrieved is not of the specified data_class
        """
        # handle special cases
        not_in_sample = ValueError(f"{data_keyword} not in {self.sample}.")
        wrong_class = ValueError(f"{data_keyword} is not {data_class}")
        if data_class is None:
            data_class = data.Sequence
        if isinstance(data_keyword, dict):
            return {k: self.get_data(v, data_class) for k, v in data_keyword.items()}
        elif isinstance(data_keyword, list):
            return [self.get_data(v, data_class) for v in data_keyword]
        elif isinstance(data_keyword, data.Sequence):
            if not isinstance(data_keyword, data_class):
                raise wrong_class
            return data_keyword
        elif data_keyword is None:
            return None
        elif isinstance(data_keyword, str) and data_keyword.startswith("default_"):
            data_keyword = self.defaults[data_keyword]
            if data_keyword is None:
                raise not_in_sample
        # if keyword is in sample.data, retreive Sequence object
        try:
            data_obj = self.data[data_keyword]
        except KeyError as exception:
            raise not_in_sample from exception
        if not isinstance(data_obj, data_class):
            raise wrong_class
        return data_obj

    def set_as_default(self, data_keyword, overwrite=True):
        """Set the given data keyword as the default for its data class

        It's data class is determined automatically. Only one default exists
        per data class and per Sample object.

        Parameters
        ----------
            data_keyword : str
                The data keyword to set as the default
            overwrite : bool, defaults to ``True``
                whether to overwrite a pre-existing default data keyword
        """
        data_object = self.data[data_keyword]
        if isinstance(data_object, data.Annotation):
            default_keyword = "default_annotation"
        elif isinstance(data_object, data.Profile):
            default_keyword = "default_profile"
        elif isinstance(data_object, data.SecondaryStructure):
            default_keyword = "default_structure"
        elif isinstance(data_object, data.Interactions):
            default_keyword = "default_interactions"
        elif isinstance(data_object, data.PDB):
            default_keyword = "default_pdb"
        else:
            return
        if overwrite or self.defaults[default_keyword] is None:
            self.defaults[default_keyword] = data_keyword

    def filter_interactions(
        self,
        interactions,
        metric=None,
        cmap=None,
        normalization=None,
        values=None,
        **kwargs,
    ):
        """sets coloring properties and applies filters to interactions data.

        Parameters
        ----------
            interactions : rnavigate.data.Interactions or data keyword string
                Interactions object to be filtered. If a string, value is
                replaced with self.get_data(interactions)
            metric : str, optional
                column of interactions data to be used as metric for coloring
                interactions.
                "Distance" will compute 3D distance in "pdb", defaulting to
                2'OH atom. "Distance_DMS" or "Distance_[atom id]" will use
                those atoms to compute distance.
            cmap (str | list, optional):
                sets the interactions colormap, used to color interactions
                according to metric values.
            normalization (str, optional):
                `"norm"`: extreme values in colormap are given to the extreme
                    values of interactions metric data
                `"bins"`: data are colored according to which bin they fall in
                    `values` defines bins (list, length = 2 less than cmap)
                `"min_max"`: extreme values in cmap are given to values beyond
                    minimum and maximum, defined by `values`
            values:
                behavior depends on normalization
                `"norm"`: values are not needed
                `"bins"`: list of floats containing the boundaries between bins
                    One fewer than the number of categories
                `"min_max"`: list of floats containing the minimum and maximum
            **kwargs: Other arguments are passed to interactions.filter()
        """
        # check for valid interactions data
        if interactions is None:
            return
        try:
            interactions = self.get_data(interactions, data.Interactions)
        except KeyError as exception:
            raise KeyError(f"{interactions} is not in {self.sample}") from exception

        if metric is None:
            metric = interactions.default_metric
        elif metric.startswith("Distance"):
            if len(metric.split("_")) == 2:
                metric, atom = metric.split("_")
            else:
                atom = "O2'"
            interactions.set_3d_distances(self.get_data("default_pdb"), atom)
        metric = {"metric_column": metric}
        if cmap is not None:
            metric["cmap"] = cmap
        if normalization is not None:
            metric["normalization"] = normalization
        if values is not None:
            metric["values"] = values
        interactions.metric = metric
        if "profile" in kwargs:
            kwargs["profile"] = self.data[kwargs["profile"]]
        else:
            try:
                kwargs["profile"] = self.get_data("default_profile")
            except ValueError:
                kwargs["profile"] = None
        if "structure" in kwargs:
            kwargs["structure"] = self.data[kwargs["structure"]]
        else:
            try:
                kwargs["structure"] = self.get_data("default_structure")
            except ValueError:
                kwargs["structure"] = None
        interactions.filter(**kwargs)

    def print_data_keywords(self, return_dict=False):
        """Print a nicely formatted, organized list of data keywords.

        Returns a dictionary of data keywords, organized by data type, if
        return_dict is True.
        """
        data_keywords = {
            "annotations": [],
            "profiles": [],
            "structures": [],
            "interactions": [],
            "pdbs": [],
        }
        for data_keyword, data_object in self.data.items():
            if data_keyword in self.defaults.values():
                data_keyword += " (default)"
            if isinstance(data_object, data.Annotation):
                data_keywords["annotations"].append(data_keyword)
            if isinstance(data_object, data.Profile):
                data_keywords["profiles"].append(data_keyword)
            if isinstance(data_object, data.SecondaryStructure):
                data_keywords["structures"].append(data_keyword)
            if isinstance(data_object, data.Interactions):
                data_keywords["interactions"].append(data_keyword)
            if isinstance(data_object, data.PDB):
                data_keywords["pdbs"].append(data_keyword)
        if return_dict:
            return data_keywords
        print(f"{self.sample} data keywords:")
        for k, v in data_keywords.items():
            print(f"  {k}:")
            for dkw in v:
                print(f"    {dkw}")
        print()



from abc import ABC, abstractmethod
import matplotlib.pyplot as plt
import math


class Plot(ABC):
    """Abstract base class for plots.

    Parameters
    ----------
    num_samples : int
        Number of samples to plot.
    rows : int, optional
        Number of rows in the plot.
    cols : int, optional
        Number of columns in the plot.
    **kwargs
        Keyword arguments passed to matplotlib.pyplot.subplots.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def __init__(self, num_samples, rows=None, cols=None, **kwargs):
        """Initialize the plot."""
        self.length = num_samples
        self.rows, self.columns = self.get_rows_columns(rows, cols)
        self.fig, self.axes = plt.subplots(
            self.rows, self.columns, squeeze=False, **kwargs
        )
        self.i = 0
        self.colorbars = []

    def get_ax(self, i=None):
        """Get the current axes object.

        Parameters
        ----------
        i : int, optional
            Index of the axes object to return. If None, return the current
            axes object.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object.
        """
        if i is None:
            i = self.i
        row = i // self.columns
        col = i % self.columns
        return self.axes[row, col]

    def add_colorbar_args(self, cmap):
        """Add colorbar arguments to the plot.

        Parameters
        ----------
        cmap : rnavigate.data.ScalarMappable
            Colormap object.
        """
        if cmap is None:
            return
        for colorbar in self.colorbars:
            if cmap.is_equivalent_to(colorbar):
                break
        else:
            self.colorbars.append(cmap)

    def plot_colorbars(self):
        """Plot colorbars.

        Returns
        -------
        ColorBar
            ColorBar plot object.
        """
        rows = len(self.colorbars)
        if rows == 0:
            return (None, None)
        plot = ColorBar(rows, rows=rows)
        for colorbar in self.colorbars:
            plot.plot_data(colorbar)
        plot.set_figure_size()
        return plot

    def get_rows_columns(self, rows=None, cols=None):
        """Get the number of rows and columns in the plot.

        Parameters
        ----------
        rows : int, optional
            Number of rows in the plot. If None, the number of rows is
            determined automatically.
        cols : int, optional
            Number of columns in the plot. If None, the number of columns is
            determined automatically.

        Returns
        -------
        rows : int
            Number of rows in the plot.
        cols : int
            Number of columns in the plot.
        """
        has_rows = isinstance(rows, int)
        has_cols = isinstance(cols, int)
        if has_rows and has_cols:
            return rows, cols
        elif has_rows:
            cols = math.ceil(self.length / rows)
        elif has_cols:
            rows = math.ceil(self.length / cols)
        elif self.length < 10:
            rows, cols = [
                (0, 0),
                (1, 1),
                (1, 2),
                (1, 3),
                (2, 2),
                (2, 3),
                (2, 3),
                (3, 3),
                (3, 3),
                (3, 3),
            ][self.length]
        else:
            cols = 4
            rows = math.ceil(self.length / cols)
        return rows, cols

    @abstractmethod
    def plot_data(self):
        pass

    def save(self, filename):
        """Saves the figure to a file

        Parameters
        ----------
            filename : string
                A file path to write to. The file format is provided by this
                file extension (svg, pdf, or png).
        """
        self.fig.savefig(filename)

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=None,
        height_ax_in=None,
        height_gap_in=None,
        width_gap_in=None,
        top_in=None,
        bottom_in=None,
        left_in=None,
        right_in=None,
    ):
        """Sets figure size so that axes sizes are always consistent.

        Parameters
        ----------
        height_ax_rel : float, optional
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, optional
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, optional
            Width of the axes in inches.
        height_ax_in : float, optional
            Height of the axes in inches.
        height_gap_in : float, optional
            Height of the gap between axes in inches.
        width_gap_in : float, optional
            Width of the gap between axes in inches.
        top_in : float, optional
            Top margin in inches.
        bottom_in : float, optional
            Bottom margin in inches.
        left_in : float, optional
            Left margin in inches.
        right_in : float, optional
            Right margin in inches.
        """
        fig = self.fig
        ax = self.axes[0, 0]
        rows = self.rows
        cols = self.columns

        try:
            right_ax = ax.get_rmax()
            top_ax = right_ax
            left_ax = right_ax * -1
            bottom_ax = left_ax
        except AttributeError:
            left_ax, right_ax = ax.get_xlim()
            bottom_ax, top_ax = ax.get_ylim()

        if width_ax_in is None:
            # width of axes in inches
            width_ax_in = (right_ax - left_ax) * width_ax_rel
        if width_gap_in is None:
            # get width from relative width * ax width
            width_gap_in = fig.subplotpars.wspace * width_ax_in
        else:
            # set relative width to gap:ax ratio
            fig.subplots_adjust(wspace=width_gap_in / width_ax_in)
        # comput subplot width
        width_subplot_in = width_gap_in * (cols - 1) + width_ax_in * cols
        if right_in is not None and left_in is not None:
            width_fig_in = left_in + width_subplot_in + right_in
            fig.subplots_adjust(
                right=(1 - (right_in / width_fig_in)), left=(left_in / width_fig_in)
            )
        elif right_in is None and left_in is None:
            right_rel = fig.subplotpars.right
            left_rel = fig.subplotpars.left
            width_fig_in = width_subplot_in / (right_rel - left_rel)
        else:
            raise ValueError("Must provide both right and left margins or neither")

        # repeat the process for figure height
        if height_ax_in is None:
            height_ax_in = (top_ax - bottom_ax) * height_ax_rel
        if height_gap_in is None:
            height_gap_in = fig.subplotpars.hspace * height_ax_in
        else:
            fig.subplots_adjust(hspace=height_gap_in / height_ax_in)
        height_subplot_in = height_gap_in * (rows - 1) + height_ax_in * rows
        if top_in is not None and bottom_in is not None:
            height_fig_in = bottom_in + height_subplot_in + top_in
            fig.subplots_adjust(
                top=(1 - (top_in / height_fig_in)), bottom=(bottom_in / height_fig_in)
            )
        elif top_in is None and bottom_in is None:
            top_rel = fig.subplotpars.top
            bottom_rel = fig.subplotpars.bottom
            height_fig_in = height_subplot_in / (top_rel - bottom_rel)
        else:
            raise ValueError("Must provide both top and bottom margins or neither")

        fig.set_size_inches(width_fig_in, height_fig_in)


class ColorBar(Plot):
    """Plot a colorbar.

    Parameters
    ----------
    rows : int, optional
        Number of rows in the plot.
    **kwargs
        Keyword arguments passed to matplotlib.pyplot.subplots.

    Attributes
    ----------
    fig : matplotlib.figure.Figure
        Figure object.
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects.
    i : int
        Index of the current plot.
    """

    def plot_data(self, colorbar):
        """Add a colorbar to the current axes.

        Parameters
        ----------
        colorbar : rnavigate.data.ScalarMappable
            Colormap object.
        """
        ax = self.get_ax(self.i)
        cax = plt.colorbar(
            colorbar,
            cax=ax,
            orientation="horizontal",
            aspect=40,
            spacing="proportional",
            **colorbar.cbar_args
        )
        if colorbar.tick_labels is not None:
            ax.set_xticklabels(colorbar.tick_labels)
        ax.set_title(colorbar.title)
        cax.outline.set_visible(False)
        cax.set_alpha(0.7)
        self.i += 1
        return (2, self.rows / 2)

    def set_figure_size(
        self,
        height_ax_rel=None,
        width_ax_rel=None,
        width_ax_in=3,
        height_ax_in=0.1,
        height_gap_in=0.75,
        width_gap_in=0.5,
        top_in=None,
        bottom_in=None,
        left_in=None,
        right_in=None,
    ):
        """Set the figure size.

        Parameters
        ----------
        height_ax_rel : float, optional
            Height of the axes relative to the y-axis limits.
        width_ax_rel : float, optional
            Width of the axes relative to the x-axis limits.
        width_ax_in : float, defaults to 3
            Width of the axes in inches.
        height_ax_in : float, defaults to 0.1
            Height of the axes in inches.
        height_gap_in : float, defaults to 0.75
            Height of the gap between axes in inches.
        width_gap_in : float, defaults to 0.5
            Width of the gap between axes in inches.
        top_in : float, optional
            Top margin in inches.
        bottom_in : float, optional
            Bottom margin in inches.
        left_in : float, optional
            Left margin in inches.
        right_in : float, optional
            Right margin in inches.
        """
        return super().set_figure_size(
            height_ax_rel,
            width_ax_rel,
            width_ax_in,
            height_ax_in,
            height_gap_in,
            width_gap_in,
            top_in,
            bottom_in,
            left_in,
            right_in,
        )



"""annotations.py contains Annotations and subclasses."""
import re
import pandas as pd
from rnavigate import data


class Annotation(data.Sequence):
    """Basic annotation class to store 1D features of an RNA sequence

    Each feature type must be a seperate instance. Feature types include:
        a group of separted nucleotides (e.g. binding pocket)
        regions of interest (e.g. coding sequence, Alu elements)
        sites of interest (e.g. m6A locations)
        primer binding sites.

    Parameters
    ----------
    input_data : list
        List will be treated according to `annotation_type` argument.
        Expected behaviors for each value of `annotation_type`:
        "sites" or "group": 1-indexed location of sites of interest
            example: [1, 10, 20, 30] is four sites, 1, 10, 20, and 30
        "spans": 1-indexed, inclusive locations of spans of interest
            example: [[1, 10], [20, 30]] is two spans, 1 to 10 and 20 to 30
        "primers": Similar to spans, but 5'/3' direction is preserved.
            example: [[1, 10], [30, 20]] forward 1 to 10, reverse 30 to 20
    annotation_type : "group", "sites", "spans", or "primers"
        The type of annotation.
    sequence : str or pandas.DataFrame
        Nucleotide sequence, path to fasta file, or dataframe containing a
        "Sequence" column.
    name : str, defaults to None
        Name of annotation.
    color : matplotlib color-like, defaults to "blue"
        Color to be used for displaying this annotation on plots.

    Attributes
    ----------
    data : pandas.DataFrame
        Stores the list of sites or regions
    name : str
        The label for this annotation for use on plots
    color : valid matplotlib color
        Color to represent annotation on plots
    sequence : str
        The reference sequence string
    """

    def __init__(self, input_data, annotation_type, sequence, name=None, color="blue"):
        """Create an Annotation."""
        super().__init__(sequence, name=name)
        self.color = color

        # make sure input data matches expected format
        valid_types = ["sites", "spans", "group", "primers"]
        self.annotation_type = annotation_type
        if isinstance(input_data, pd.DataFrame):
            self.data = input_data
        elif annotation_type in ["spans", "primers"]:
            self.data = self.from_spans(input_data)
        elif annotation_type in ["sites", "group"]:
            self.data = self.from_sites(input_data)
        else:
            raise ValueError(f"annotation_type not in {valid_types}")

    @classmethod
    def from_boolean_array(
        cls, values, sequence, annotation_type, name, color="blue", window=1
    ):
        """Create an Annotation from an array of boolean values.

        True values are used to create the Annotation.

        Parameters
        ----------
        values : list of True or False
            the boolean array
        sequence : string or rnav.data.Sequence
            the sequence of the Annotation
        annotation_type : "spans", "sites", "primers", or "group"
            the type of the new annotation
            If "spans" or "primers", adjacent True values, or values within
            window are collapse to a region.
        name : string
            a name for labelling the annotation.
        color : string, defaults to "blue"
            a color for plotting the annotation
        window : integer, defaults to 1
            a window around True values to include in the annotation.

        Returns
        -------
        rnavigate.data.Annotation
            the new Annotation
        """
        annotations = []
        current_annotation = None
        pad = window // 2
        for i, value in enumerate(values):
            position = i + 1
            if value and annotation_type in ["sites", "group"]:
                annotations.append(position)
            elif value and annotation_type == "spans":
                start = max(1, position - pad)
                stop = min(len(sequence), position + pad)
                if current_annotation is None:
                    current_annotation = [start, stop]
                elif start <= current_annotation[1] + 1:
                    current_annotation[1] = stop
                elif start > current_annotation[1]:
                    annotations.append(current_annotation)
                    current_annotation = [start, stop]
        if annotation_type == "spans":
            annotations.append(current_annotation)
        return cls(
            input_data=annotations,
            annotation_type=annotation_type,
            color=color,
            sequence=sequence,
            name=name,
        )

    def from_spans(self, spans):
        """Create the self.data dataframe from a list of spans."""
        data_dict = {"start": [], "end": []}
        for span in spans:
            if (len(span) != 2) and any(not isinstance(pos, int) for pos in span):
                raise ValueError(
                    f"{self.annotation_type} must be a list of pairs of "
                    f"integers:\n{spans}"
                )
            start, end = span
            data_dict["start"].append(int(start))
            data_dict["end"].append(int(end))
        return pd.DataFrame(data_dict)

    def from_sites(self, sites):
        """Create the self.data dataframe from a list of sites."""
        if any(not isinstance(site, int) for site in sites):
            raise ValueError(
                f"{self.annotation_type} must be a list of integers:\n{sites}"
            )
        return pd.DataFrame({"site": sites})

    def get_aligned_data(self, alignment):
        """Aligns this Annotation to a new sequence and returns a copy.

        Parameters
        ----------
        alignment : rnavigate.data.Alignment
            Alignment object used to align to a new sequence.

        Returns
        -------
        rnavigate.data.Annotation
            A new Annotation with the same name, color, and annotation
            type, but with the input data aligned to the target sequence.
        """
        new_input_data = alignment.map_dataframe(self.data, self.data.columns)
        return Annotation(
            name=self.name,
            color=self.color,
            sequence=alignment.target_sequence,
            annotation_type=self.annotation_type,
            input_data=new_input_data,
        )

    def get_sites(self):
        """Returns a list of nucleotide positions included in this annotation.

        Returns
        -------
        sites : tuple
            a list of nucleotide positions
        """
        if self.annotation_type in ["spans", "primers"]:
            sites = []
            for _, row in self.data.iterrows():
                sites.extend(list(range(row["start"], row["end"] + 1)))
        elif self.annotation_type in ["sites", "group"]:
            sites = list(self.data["site"].values)
        return sites

    def get_subsequences(self, buffer=0):
        subsequences = []
        if self.annotation_type in ["spans", "primers"]:
            for _, (start, stop) in self.data[["start", "stop"]].iterrows():
                subsequences.append(self.sequence[start - 1 - buffer : stop + buffer])
        elif self.annotation_type in ["sites", "groups"]:
            for _, (site) in self.data[["sites"]].iterrows():
                subsequences.append(self.sequence[site - 1 - buffer : stop + buffer])
        return subsequences

    def __getitem__(self, idx):
        return self.data.loc[idx]

    def __iter__(self):
        for _, row in self.data.iterrows():
            yield row

    def __len__(self):
        return len(self._list)


class Motif(Annotation):
    """Automatically annotates the occurances of a sequence motif as spans.

    Parameters
    ----------
    input_data : str
        sequence motif to search for.
        Uses conventional nucleotide codes.
        e.g. "DRACH" = [AGTU] [AG] A C [ATUC]
    sequence : str or pandas.DataFrame
        Nucleotide sequence, path to fasta file, or dataframe containing a
        "Sequence" column.
    name : str, defaults to None
        Name of annotation.
    color : matplotlib color-like, defaults to "blue"
        Color to be used for displaying this annotation on plots.

    Attributes
    ----------
    data : pandas.DataFrame
        Stores the list of regions that match the motif
    name : str
        The label for this annotation for use on plots
    color : valid matplotlib color
        Color to represent annotation on plots
    sequence : str
        The reference sequence string
    """

    def __init__(self, input_data, sequence, name=None, color="blue"):
        """Creates a Motif annotation"""
        self.motif = input_data
        span_list = self.get_spans_from_motif(sequence, input_data)
        super().__init__(
            name=name,
            sequence=sequence,
            annotation_type="spans",
            input_data=span_list,
            color=color,
        )

    def get_spans_from_motif(self, sequence, motif):
        """Returns a list of spans for each location of motif found within sequence.

        Parameters
        ----------
        sequence : string
            sequence to be searched
        motif : string
            sequence motif to searched for.

        Returns
        -------
        spans : list of lists
            list of [start, end] positions of each motif occurance
        """
        sequence = data.normalize_sequence(sequence=sequence)
        nuc_codes = {
            "A": "A",
            "T": "T",
            "U": "U",
            "G": "G",
            "C": "C",
            "B": "[CGTU]",
            "D": "[ATUG]",
            "H": "[ATUC]",
            "V": "[ACG]",
            "W": "[ATU]",
            "S": "[CG]",  # strong and weak
            "M": "[AC]",
            "K": "[GTU]",  # amino and ketone
            "R": "[AG]",
            "Y": "[CTU]",  # purine and pyrimidine
            "N": "[ATUGC]",
        }  # any nuc
        re_pattern = "".join([nuc_codes[n] for n in motif])
        spans = []
        for match in re.finditer(re_pattern, sequence):
            start, end = match.span()
            spans.append([start + 1, end])
        return spans

    def get_aligned_data(self, alignment):
        """Searches the new sequence for the motif and returns a new Motif annotation.

        Parameters
        ----------
        alignment : rnavigate.data.Alignment
            Alignment object used to align to a new sequence.

        Returns
        -------
        rnavigate.data.Motif
            A new Motif with the same name, color, and motif
            but with the input data aligned to the target sequence.
        """
        return Motif(
            input_data=self.motif,
            name=self.name,
            color=self.color,
            sequence=alignment.target_sequence,
        )


class ORFs(Annotation):
    """Automatically annotations occurances of open-reading frames as spans.

    Parameters
    ----------
    input_data : "longest" or "all"
        which ORFs to annotate. "longest" annotates the longest ORF. "all"
        annotates all potential ORFs.
    sequence : str or pandas.DataFrame
        Nucleotide sequence, path to fasta file, or dataframe containing a
        "Sequence" column.
    name : str, defaults to None
        Name of annotation.
    color : matplotlib color-like, defaults to "blue"
        Color to be used for displaying this annotation on plots.

    Attributes
    ----------
    data : pandas.DataFrame
        Stores the list of regions that match the motif
    name : str
        The label for this annotation for use on plots
    color : valid matplotlib color
        Color to represent annotation on plots
    sequence : str
        The reference sequence string
    """

    def __init__(self, input_data, name=None, sequence=None, color="blue"):
        """Creates an ORF annotation"""
        self.input_data = input_data
        span_list = self.get_spans_from_orf(sequence, which=input_data)
        super().__init__(
            name=name,
            sequence=sequence,
            annotation_type="spans",
            input_data=span_list,
            color=color,
        )

    def get_spans_from_orf(self, sequence, which="all"):
        """Given a sequence string, returns spans for specified ORFs

        Parameters
        ----------
        sequence : string
            RNA nucleotide sequence
        which : "longest" or "all", defaults to "all"
            "all" returns all spans, "longest" returns the longest span

        Returns
        -------
        list of tuples
            (start, end) position of each ORF 1-indexed, inclusive
        """
        if isinstance(sequence, data.Sequence):
            sequence = sequence.sequence
        sequence = sequence.upper().replace("T", "U")
        spans = []
        stop_codons = "UAA|UAG|UGA"
        start_codon = "AUG"
        stop_sites = [m.end() for m in re.finditer(stop_codons, sequence)]
        start_sites = [m.start() + 1 for m in re.finditer(start_codon, sequence)]
        for start in start_sites:
            for stop in stop_sites:
                if (stop - start) % 3 == 2 and start < stop:
                    spans.append([start, stop])
        if which == "all":
            return spans
        if which == "longest":
            lengths = [end - start for start, end in spans]
            index = lengths.index(max(lengths))
            return [spans[index]]

    def get_aligned_data(self, alignment):
        """Searches the new sequence for ORFs and returns a new ORF annotation.

        Parameters
        ----------
        alignment : rnavigate.data.Alignment
            Alignment object used to align to a new sequence.

        Returns
        -------
        rnavigate.data.ORFs
            A new ORFs annotation with the same name, color, and input_data
            but with the input data aligned to the target sequence.
        """
        return ORFs(
            input_data=self.input_data,
            name=self.name,
            color=self.color,
            sequence=alignment.target_sequence,
        )


def domains(input_data, names, colors, sequence):
    """Create a list of Annotations from a list of spans.

    Currently, domains functionality in RNAvigate just uses a list of spans. In the
    future, this should be a dedicated class. Generally, domains should cover an entire
    sequence without overlap, but this is not enforced.
    e.g. [[1, 100], [101, 200]] for a 200 nt sequence.

    Parameters
    ----------
    input_data : list of lists
        list of spans for each domain
    names : list of strings
        list of names for each domain
    colors : list of valid matplotlib colors
        list of colors for each domain
    sequence : string
        sequence to be annotated

    Returns
    -------
    list of rnavigate.data.Annotation
        list of Annotations
    """
    return [
        Annotation(
            input_data=[span],
            annotation_type="spans",
            name=name,
            color=color,
            sequence=sequence,
        )
        for span, name, color in zip(input_data, names, colors)
    ]



"""Module for plotting arc plots."""

from rnavigate import plots, styles


class AP(plots.Plot):
    """Class for plotting arc plots

    Parameters
    ----------
    num_samples : int
        Number of samples to plot
    nt_length : int
        Length of the sequence
    region : tuple of 2 integers, optional
        starting and ending positions of the region to plot.
        Default is "all", which plots the entire sequence.
    track_labels : bool, optional
        Whether to plot track labels. Default is True.
    **kwargs
        Additional keyword arguments are passed to plots.Plot

    Attributes
    ----------
    nt_length : int
        Length of the sequence
    region : tuple of 2 integers
        starting and ending positions of the region to plot.
    track_labels : bool
        Whether to plot track labels.
    fig : matplotlib.figure.Figure
        Figure object containing the plot
    axes : numpy.ndarray of matplotlib.axes.Axes
        Array of axes objects containing the plots
    i : int
        Index of the current plot
    """

    def __init__(
        self, num_samples, nt_length, region="all", track_labels=True, **kwargs
    ):
        """Initialize AP object."""
        self.track_labels = track_labels
        if region == "all":
            self.nt_length = nt_length
            self.region = (1, nt_length)
        else:
            self.nt_length = region[1] - region[0] + 1
            self.region = region
        super().__init__(num_samples, **kwargs)

    def plot_data(
        self,
        sequence,
        structure=None,
        structure2=None,
        interactions=None,
        interactions2=None,
        profile=None,
        annotations=None,
        domains=None,
        label="",
        ax=None,
        seqbar=True,
        title=True,
        panels=None,
        annotation_mode="track",
        track_height=None,
        profile_scale_factor=1,
        plot_error=False,
        nt_ticks=(20, 5),
    ):
        """Add data to the next (or specified) plot axes.

        This function assumes data has already been aligned to a common sequence.
        rnavigate.plot_ functions can be used to automatically align data.

        Parameters
        ----------
        sequence : rnavigate.data.Sequence
            Sequence object containing the sequence to plot
        structure : rnavigate.data.Structure, optional
            Structure object containing a structure to plot
        structure2 : rnavigate.data.Structure, optional
            Structure object containing a structure to compare to the first
        interactions : rnavigate.data.Interactions, optional
            Interactions object containing inter-nucleotide data to plot
        interactions2 : rnavigate.data.Interactions, optional
            Interactions object containing other inter-nucleotide data to plot
        profile : rnavigate.data.Profile, optional
            Profile object containing per-nucleotide data to plot
        annotations : list of rnavigate.data.Annotation, optional
            List of Annotation objects containing annotations to plot
        domains : list of rnavigate.data.Spans, optional
            List of Spans objects containing domains to plot
        label : str, defaults to ""
            Label for the title of the plot.
        ax : matplotlib.axes.Axes, optional
            Axes object to plot on. If None, the next axes in the figure will be used.
        seqbar : bool, Defaults to True
            Whether to plot the sequence track.
        title : bool, defaults to True
            Whether to show the title.
        panels : dict, optional
            Dictionary of panels to plot, with keys being the panel name and values
            being the panel location. Default is {"interactions": "bottom",
            "interactions2": "bottom", "structure": "top", "profile": "top"}
        annotation_mode : "track" or "bar", defaults to "track"
            Mode for plotting annotations.
        track_height : int, optional
            Height of the track. If None, the height is automatically determined.
        profile_scale_factor : float, defaults to 1
            Scale factor for the profile track.
        plot_error : bool, defaults to False
            Whether to plot the error bars for the profile track.
        nt_ticks : tuple of 2 ints, optional
            Major and minor tick spacing for the nucleotide axis. Default is (20, 5).
        """
        ax = self.get_ax(ax)
        if panels is None:
            panels = {}
        panels = {
            "interactions": "bottom",
            "interactions2": "bottom",
            "structure": "top",
            "profile": "top",
        } | panels
        if annotations is None:
            annotations = []

        annotation_gap = 4 * len(annotations) * (annotation_mode == "track")
        seqbar_height = 2 * seqbar
        domains_height = 6 * (domains is not None)
        if track_height is None:
            track_height = annotation_gap + seqbar_height + domains_height

        yvalues = {"bottom": 0, "top": track_height}
        if structure is not None:
            structure = structure.as_interactions(structure2)
            plots.plot_interactions_arcs(
                ax=ax,
                interactions=structure,
                panel=panels["structure"],
                yvalue=yvalues[panels["structure"]],
                region=self.region,
            )
            self.add_colorbar_args(structure.cmap)
        if interactions is not None:
            plots.plot_interactions_arcs(
                ax=ax,
                interactions=interactions,
                panel=panels["interactions"],
                yvalue=yvalues[panels["interactions"]],
                region=self.region,
            )
            self.add_colorbar_args(interactions.cmap)
        if interactions2 is not None:
            plots.plot_interactions_arcs(
                ax=ax,
                interactions=interactions2,
                panel=panels["interactions2"],
                yvalue=yvalues[panels["interactions2"]],
                region=self.region,
            )
            self.add_colorbar_args(interactions2.cmap)
        if profile is not None:
            if panels["profile"] == "bottom":
                scale_factor = profile_scale_factor * -1
                bottom = 0
            else:
                scale_factor = profile_scale_factor
                bottom = track_height
            plots.plot_profile_bars(
                ax=ax,
                profile=profile,
                bottom=bottom,
                scale_factor=scale_factor,
                plot_error=plot_error,
                region=self.region,
            )
            self.add_colorbar_args(profile.cmap)

        yticks, ylabels = [], []
        if seqbar:
            plots.plot_sequence_track(
                ax,
                sequence.sequence,
                yvalue=0,
                height=2,
                ytrans="data",
                region=self.region,
            )
            yticks.append(1)
            ylabels.append("sequence")
            self.add_colorbar_args(styles.get_nt_cmap())
        if domains is not None:
            for domain in domains:
                plots.plot_domain_track(
                    ax=ax,
                    spans=domain,
                    yvalue=seqbar_height,
                    height=6,
                    region=self.region,
                )
            yticks.append(seqbar_height + 3)
            ylabels.append("domains")
        for i, annotation in enumerate(annotations):
            yvalue = seqbar_height + domains_height + 4 * (i + 1) - 2
            plots.plot_annotation_track(
                ax,
                annotation=annotation,
                yvalue=yvalue,
                height=4,
                mode=annotation_mode,
                region=self.region,
            )
            yticks.append(yvalue)
            ylabels.append(annotation.name)
        if not self.track_labels:
            yticks, ylabels = [], []
        self.set_axis(
            ax=ax,
            sequence=sequence.sequence,
            track_height=track_height,
            nt_ticks=nt_ticks,
            yticks=yticks,
            ylabels=ylabels,
        )
        if title:
            ax.set_title(label)
        self.i += 1

    def set_axis(
        self,
        ax,
        sequence,
        track_height=0,
        nt_ticks=(20, 5),
        max_height=300,
        yticks=None,
        ylabels=None,
    ):
        """Set up the plotting axis settings for an aesthetic arc plot.

        Sets the following properties of the given axis:
        1. spine positions
        2. x-axis and y-axis limits
        3. x-axis tick labels and positions according to `sequence`
        4. background boxes for x-axis tick labels
        5. y-axis tick labels and positions according to `track_height` and `ylabels`

        Parameters
        ----------
        ax : matplotlib.axes.Axes
            Axes object to set up.
        sequence : str
            Sequence to plot.
        track_height : int, optional
            Height of the track. If None, the height is automatically determined.
        nt_ticks : tuple of 2 ints, optional
            Major and minor tick spacing for the nucleotide axis. Default is (20, 5).
        max_height : int, optional
            Maximum height of the plot. Default is 300.
        yticks : list of ints, optional
            List of ytick positions. If None, the yticks are automatically determined.
        ylabels : list of str, optional
            List of ytick labels. If None, the ylabels are automatically determined.
        """
        ax.spines["left"].set_color("none")
        ax.spines["right"].set_color("none")
        ax.spines["bottom"].set(position=("data", 0), visible=False)
        ax.spines["top"].set_color("none")
        height = min(max_height, self.nt_length / 2)
        mn, mx = self.region
        ax.set(
            xlim=(mn - 0.5, mx + 0.5),
            ylim=(-height - 1, height + 1 + track_height),
            yticks=yticks,
            yticklabels=ylabels,
            axisbelow=False,
        )
        plots.set_nt_ticks(
            ax=ax,
            sequence=sequence,
            region=self.region,
            major=nt_ticks[0],
            minor=nt_ticks[1],
        )
        for label in ax.get_xticklabels():
            label.set_bbox(
                {
                    "facecolor": "white",
                    "edgecolor": "None",
                    "alpha": 0.5,
                    "boxstyle": "round,pad=0.1,rounding_size=0.2",
                }
            )

    def set_figure_size(
        self,
        height_ax_rel=0.03,
        width_ax_rel=0.03,
        width_ax_in=None,
        height_ax_in=None,
        height_gap_in=0.5,
        width_gap_in=0.5,
        top_in=1,
        bottom_in=1,
        left_in=1,
        right_in=1,
    ):
        """Set the figure size for an arc plot.

        Parameters
        ----------
        height_ax_rel : float, Default is 0.03.
            Relative height of each axes.
        width_ax_rel : float, Default is 0.03.
            Relative width of each axes.
        width_ax_in : float, optional
            Absolute width of each axes in inches. If None, the width is automatically
            determined.
        height_ax_in : float, optional
            Absolute height of each axes in inches. If None, the height is
            automatically determined.
        height_gap_in : float, Default is 0.5.
            Absolute height of the gap between axes in inches.
        width_gap_in : float, Default is 0.5.
            Absolute width of the gap between axes in inches.
        top_in : float, Default is 1.
            Absolute top margin in inches.
        bottom_in : float, Default is 1.
            Absolute bottom margin in inches.
        left_in : float, Default is 1.
            Absolute left margin in inches.
        right_in : float, Default is 1.
            Absolute right margin in inches.
        """
        super().set_figure_size(
            height_ax_rel=height_ax_rel,
            width_ax_rel=width_ax_rel,
            width_ax_in=width_ax_in,
            height_ax_in=height_ax_in,
            height_gap_in=height_gap_in,
            width_gap_in=width_gap_in + self.track_labels * 0.5,
            top_in=top_in,
            bottom_in=bottom_in,
            left_in=left_in,
            right_in=right_in,
        )



import numpy as np
import matplotlib.patches as mp_patches
import matplotlib.collections as mp_collections
import matplotlib.colors as mp_colors
from rnavigate import data, plots


def get_contrasting_colors(colors):
    """Get contrasting colors for a list of colors.

    Returns a list of "k" (black) or "w" (white) for each color in the input list,
    which ever contrasts better with the input color.

    Parameters
    ----------
    colors : list of str
        List of colors to get contrasting colors for.

    Returns
    -------
    list of str
        List of "k" or "w" for each color in the input list.
    """
    contrasting = ["k"] * len(colors)
    for i, color in enumerate(colors):
        r, g, b = mp_colors.to_rgb(color)
        if (r * 0.299 + g * 0.587 + b * 0.114) < 175 / 256:
            contrasting[i] = "w"
    return np.array(contrasting)


def adjust_spines(ax, spines):
    """Places the given spines on the given axis, removes others.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to adjust the spines of.
    spines : list of str
        The spines to adjust. Valid options are "left", "right", "top", and "bottom".
    """
    for loc, spine in ax.spines.items():
        if loc in spines:
            spine.set_position(("outward", 10))  # outward by 10 points
        else:
            spine.set_color("none")  # don't draw spine
    if "left" in spines:
        ax.yaxis.set_ticks_position("left")
    else:
        ax.yaxis.set_ticks([])
    if "bottom" in spines:
        ax.xaxis.set_ticks_position("bottom")
    else:
        ax.xaxis.set_ticks([])


def clip_spines(ax, spines):
    """Clips the given spines to the range of the ticks.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to clip the spines of.
    spines : list of str
        The spines to clip. Valid options are "left", "right", "top", and "bottom".
    """
    for spine in spines:
        if spine in ["left", "right"]:
            ticks = ax.get_yticks()
        if spine in ["top", "bottom"]:
            ticks = ax.get_xticks()
        ax.spines[spine].set_bounds((min(ticks), max(ticks)))


def get_nt_ticks(sequence, region, gap):
    """Get the ticks and labels for a region of a sequence.

    Dashes are ignored when counting nucleotide positions.

    Parameters
    ----------
    sequence : str or rnavigate.data.Sequence
        The sequence to get ticks and labels for.
    region : tuple of int
        The region of the sequence to get ticks and labels for.
    gap : int
        The gap between major ticks.

    Returns
    -------
    ticks, labels : tuple of list of int
        The ticks and labels for the given region of the sequence.
    """
    if isinstance(sequence, data.Sequence):
        sequence = sequence.sequence
    start, end = region
    labels = []
    ticks = []
    pos = 0
    for i, nt in enumerate(sequence):
        valid = nt != "-"
        if valid:
            pos += 1
        if valid & (pos % gap == 0 or pos == start) and start <= pos <= end:
            labels.append(pos)
            ticks.append(i + 1)
    return ticks, labels


def set_nt_ticks(ax, sequence, region, major, minor):
    """Set the ticks and labels for a region of a sequence.

    Dashes are ignored when counting nucleotide positions.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to set the ticks and labels of.
    sequence : str or rnavigate.data.Sequence
        The sequence to set ticks and labels for.
    region : tuple of int
        The region of the sequence to set ticks and labels for.
    major : int
        The gap between major ticks.
    minor : int
        The gap between minor ticks.
    """
    ticks, labels = get_nt_ticks(sequence, region, major)
    ax.set_xticks(ticks=ticks)
    ax.set_xticklabels(labels=labels)
    ax.set_xticks(minor=True, ticks=get_nt_ticks(sequence, region, minor)[0])


def box_xtick_labels(ax):
    """Place a transparent box behind the xtick labels of the provided axis."""
    for label in ax.get_xticklabels():
        label.set_bbox(
            {
                "facecolor": "white",
                "edgecolor": "None",
                "alpha": 0.5,
                "boxstyle": "round,pad=0.1,rounding_size=0.2",
            }
        )


def plot_sequence_alignment(ax, alignment, labels, top=5, bottom=-5, ytrans="data"):
    """Plot a sequence alignment.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the alignment on.
    alignment : rnavigate.data.Alignment
        The alignment to plot.
    labels : tuple of str
        The labels for the two sequences in the alignment.
    top : int, optional
        The y-value to plot the top sequence at.
    bottom : int, optional
        The y-value to plot the bottom sequence at.
    ytrans : str, optional
        The transform to use for the y-values. Valid options are "data" and "axes".
    """
    al1 = alignment.alignment1
    al2 = alignment.alignment2
    height = (top - bottom) / 3
    plots.plot_sequence_track(
        ax, sequence=al1, yvalue=bottom, height=height, ytrans=ytrans
    )
    plots.plot_sequence_track(
        ax,
        sequence=al2,
        yvalue=top,
        height=height,
        ytrans=ytrans,
        verticalalignment="top",
    )
    set_nt_ticks(ax=ax, sequence=al1, region=(1, len(al1)), major=20, minor=5)
    ax.set_xlabel(labels[0], loc="left")
    ax.spines["bottom"].set(position=(ytrans, bottom), visible=False)
    ax2 = ax.twiny()
    ax2.set(xlim=ax.get_xlim())
    ax2.spines["top"].set(position=(ytrans, top), visible=False)
    set_nt_ticks(ax=ax2, sequence=al2, region=(1, len(al2)), major=20, minor=5)
    ax2.set_xlabel(labels[1], loc="left")

    for spine in ["top", "bottom", "left", "right"]:
        ax2.spines[spine].set_visible(False)
    for spine in ["top", "bottom", "left", "right"]:
        ax.spines[spine].set_visible(False)

    for idx, (nt1, nt2) in enumerate(zip(al1, al2)):
        if nt1.upper().replace("T", "U") == nt2.upper().replace("T", "U"):
            ax.fill_between(
                x=[idx + 0.5, idx + 1.5],
                y1=bottom + height,
                y2=top - height,
                color="grey",
                ec="none",
            )


def plot_interactions_arcs(ax, interactions, panel, yvalue=0, region="all"):
    """Plot interactions as arcs.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the interactions on.
    interactions : rnavigate.data.Interactions
        The interactions to plot.
    panel : "top" or "bottom"
        The panel to plot the interactions on.
    yvalue : float, optional
        The y-value at which to plot the interactions.
    region : tuple of int, optional
        The region of the sequence to plot interactions for.
    """
    if region == "all":
        region = [1, interactions.length]
    mn, mx = region
    ij_colors = interactions.get_ij_colors()
    patch_list = []
    for i, j, color in zip(*ij_colors):
        if j < i:  # flip the order
            i, j = j, i
        if not (mn < i < mx) and not (mn < j < mx):
            continue
        center = ((i + j) / 2.0, yvalue)
        if panel == "top":
            theta1, theta2 = 0, 180
        elif panel == "bottom":
            theta1, theta2 = 180, 360
        radius = 0.5 + (j - i) / 2.0
        patch_list.append(
            mp_patches.Wedge(
                center, radius, theta1, theta2, color=color, width=1, ec="none"
            )
        )
    ax.add_collection(mp_collections.PatchCollection(patch_list, match_original=True))


def plot_profile_bars(
    ax, profile, scale_factor=1, plot_error=True, bottom=0, region="all"
):
    """Plot per-nucleotide data as colored bars.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the profile on.
    profile : rnavigate.data.Profile
        The profile to plot.
    scale_factor : float, optional
        The factor by which to multiply per-nucleotide values.
    plot_error : bool, optional
        Whether to plot error bars.
    bottom : float, optional
        The y-value at which to plot the bars.
    region : tuple of int, optional
        The region of the sequence to plot bars for.
    """
    if region == "all":
        region = [1, profile.length]
    mn, mx = region
    data = profile.get_plotting_dataframe()
    values = data["Values"]
    colormap = data["Colors"]
    nts = data["Nucleotide"]
    if plot_error and ("Errors" in data.columns):
        yerr = data["Errors"] * scale_factor
        ax.bar(
            nts[mn - 1 : mx],
            values[mn - 1 : mx] * scale_factor,
            align="center",
            bottom=bottom,
            width=1,
            color=colormap[mn - 1 : mx],
            edgecolor=colormap[mn - 1 : mx],
            linewidth=0.0,
            yerr=yerr[mn - 1 : mx],
            ecolor=(0, 0, 1 / 255.0),
            capsize=0,
        )
    else:
        ax.bar(
            nts[mn - 1 : mx],
            values[mn - 1 : mx] * scale_factor,
            align="center",
            width=1,
            color=colormap[mn - 1 : mx],
            bottom=bottom,
            edgecolor=colormap[mn - 1 : mx],
            linewidth=0.0,
        )


def plot_profile_skyline(ax, profile, label, columns, errors):
    """Plot per-nucleotide data as a skyline plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the profile on.
    profile : rnavigate.data.Profile
        The profile to plot.
    label : str
        The label to use for the plot legend.
    columns : list of str
        The columns of the profile to plot.
    errors : list of str
        The columns of the profile to use for error bars.
    """
    values = profile.data
    if columns is None:
        columns = profile.metric
    if isinstance(columns, str) and isinstance(errors, (str, type(None))):
        columns = [columns]
        errors = [errors]
    if errors is None:
        errors = [None] * len(columns)
    if len(errors) != len(columns):
        raise ValueError("columns and errors lists must be the same length")
    for column, error in zip(columns, errors):
        lines = ax.plot(
            values["Nucleotide"],
            values[column],
            drawstyle="steps-mid",
            label=f"{label}: {column.replace('_', ' ')}",
        )
        if error is not None:
            ax.fill_between(
                values["Nucleotide"],
                values[column] - values[error],
                values[column] - values[error],
                step="mid",
                color=lines[0].get_color(),
                alpha=0.25,
                lw=0,
            )



from rnavigate.analysis.logcompare import LogCompare
from rnavigate.analysis.lowss import LowSS
from rnavigate.analysis.deltashape import DeltaSHAPE, DeltaSHAPEProfile
from rnavigate.analysis.fragmapper import FragMaP, Fragmapper
from rnavigate.analysis.check_sequence import SequenceChecker


__all__ = [
    "LogCompare",
    "LowSS",
    "DeltaSHAPE",
    "DeltaSHAPEProfile",
    "FragMaP",
    "Fragmapper",
    "SequenceChecker",
]



import numpy as np
import matplotlib as mp
import matplotlib.patches as mp_patches
from rnavigate import styles


# 1-dimensional x-axis tracks
def plot_sequence_track(
    ax,
    sequence,
    yvalue=-0.05,
    height=0.05,
    ytrans="data",
    verticalalignment="bottom",
    region="all",
):
    """Plot a sequence track along the x-axis of a plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the sequence track on.
    sequence : str
        The sequence to plot.
    yvalue : float, defaults to -0.05
        The y-value of the sequence track.
    height : float, defaults to 0.05
        The height of the sequence track.
    ytrans : "data" or "axes", defaults to "data"
        The y-axis coordinate system.
    verticalalignment : "top" or "bottom", defaults to "bottom"
        The vertical alignment of the sequence track.
    region : list of 2 int, defaults to "all"
        Start and end positions of the region to plot. If "all", plot the entire
        sequence.
    """
    style = styles.settings["sequence_bar"]
    ymin, ymax = ax.get_ylim()
    if ytrans == "axes":
        yvalue = (ymax - ymin) * yvalue + ymin
        height = (ymax - ymin) * height
    if region == "all":
        region = [1, len(sequence)]
    sequence = sequence[region[0] - 1 : region[1]]
    if style == "alphabet":
        for i, nt in enumerate(sequence):
            # set font style and colors for each nucleotide
            font_prop = mp.font_manager.FontProperties(
                family="monospace", style="normal", weight="bold", size="4"
            )
            col = styles.get_nt_color(nt)
            ax.text(
                i + region[0],
                yvalue,
                nt,
                fontproperties=font_prop,
                color=col,
                horizontalalignment="center",
                verticalalignment=verticalalignment,
            )
    if style == "bars":
        sequence = list(sequence)
        colors = [styles.get_nt_color(nt) for nt in sequence]
        is_uppercase = np.char.isupper(sequence)
        heights = np.full(len(sequence), height / 2)
        heights *= is_uppercase + 1
        heights *= np.char.not_equal(sequence, "-")
        if verticalalignment == "top":
            heights *= -1
        x = np.arange(region[0], region[1] + 1)
        ax.bar(
            x, heights, bottom=yvalue, width=1, color=colors, ec="none", clip_on=False
        )
    ax.set_ylim(ymin, ymax)


def plot_annotation_track(
    ax,
    annotation,
    yvalue,
    height,
    mode,
    region="all",
    ytrans="data",
):
    """Plot an annotation track along the x-axis of a plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the annotation track on.
    annotation : rnavigate.data.Annotation
        The annotation to plot.
    yvalue : float
        The y-value of the annotation track.
    height : float
        The height of the annotation track.
    mode : "track" or "bar"
        The annotation mode.
    region : list of 2 int, defaults to "all"
        Start and end positions of the region to plot. If "all", plot the entire
        sequence.
    ytrans : "data" or "axes", defaults to "data"
        The y-axis coordinate system.
    """
    if region == "all":
        region = [1, annotation.length]
    mn, mx = region
    if ytrans == "axes":
        ymin, ymax = ax.get_ylim()
        yvalue = (ymax - ymin) * yvalue + ymin
        height = (ymax - ymin) * height
    color = annotation.color
    modes = ["track", "bar"]

    def plot_track(*s, alpha=1):
        start, end = min(s) - 0.5, max(s) + 0.5
        if mode == "track":
            ax.plot(
                [start, end],
                [yvalue] * 2,
                color=color,
                alpha=alpha,
                lw=5,
                solid_capstyle="butt",
                clip_on=False,
            )
        elif mode == "bar" and len(s) == 2:
            ax.axvspan(start - 0.5, end + 0.5, fc=color, ec="none", alpha=0.1)
        elif mode == "bar" and len(s) == 1:
            ax.axvline(start + 0.5, color=color, ls=":")

    assert mode in modes, f"annotation mode must be one of: {modes}"
    a_type = annotation.annotation_type
    if a_type == "spans" or (a_type == "primers" and mode == "bar"):
        for start, end in annotation:
            plot_track(start, end)
    elif a_type == "sites":
        for site in annotation.data["site"]:
            plot_track(site)
    elif a_type == "group":
        sites = annotation.data["site"]
        plot_track(min(sites), max(sites), alpha=0.5)
        for site in sites:
            plot_track(site)
    elif a_type == "primers":
        for start, end in annotation:
            if start > end:
                start, end = start + 0.5, end - 0.5
            else:
                start, end = start - 0.5, end + 0.5
            ax.arrow(
                x=start,
                y=yvalue,
                dx=end - start + 1,
                dy=0,
                color=color,
                shape="right",
                clip_on=False,
                head_width=height * 0.8,
                head_length=3,
                length_includes_head=True,
            )


def plot_domain_track(ax, spans, yvalue, height, region="all", ytrans="data"):
    """Plot a domain track along the x-axis of a plot.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes to plot the domain track on.
    spans : rnavigate.data.Spans
        The spans to plot.
    yvalue : float
        The y-value of the domain track.
    height : float
        The height of the domain track.
    region : list of 2 int, defaults to "all"
        Start and end positions of the region to plot. If "all", plot the entire
        sequence.
    ytrans : "data" or "axes", defaults to "data"
        The y-axis coordinate system.
    """
    if region == "all":
        region = [1, spans.length]
    mn, mx = region
    text_height = height
    if ytrans == "axes":
        ymin, ymax = ax.get_ylim()
        yvalue = (ymax - ymin) * yvalue + ymin
        height = (ymax - ymin) * height
        text_height = 6
    name = spans.name
    color = spans.color
    for start, end in spans:
        start = max([mn, start])
        end = min([mx, end])
        rect = mp_patches.Rectangle(
            xy=(start - 0.5, yvalue),
            width=end - start + 1,
            height=height,
            linewidth=1,
            edgecolor="black",
            facecolor=color,
            clip_on=False,
        )
        ax.add_patch(rect)
        ax.text(
            x=(end - start) / 2 + start,
            y=yvalue + height / 2,
            s=name,
            zorder=25,
            horizontalalignment="center",
            verticalalignment="center",
            fontdict={"size": text_height * 1.5},
            clip_on=False,
        )



from operator import ge, le, gt, lt, eq, ne
import pandas as pd
import matplotlib.colors as mpc
import numpy as np
import seaborn as sns
from rnavigate import data


class Interactions(data.Data):
    """A class for storing and manipulating interactions data.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing interactions data.
        If dataframe, the dataframe containing interactions data. The dataframe
        must contain columns "i", "j", and self.metric. Dataframe may also
        include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the interactions data.
    metric : string
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap)
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict
        kwargs passed to pandas.read_table() when reading input_data.
    window : int
        The window size used to generate the interactions data.
    name : str
        The name of the data object.

    Attributes
    ----------
    data : pandas.DataFrame
        The interactions data.
    window : int
        The window size that is being represented by i-j pairs.
    """

    def __init__(
        self,
        input_data,
        sequence,
        metric,
        metric_defaults,
        read_table_kw=None,
        window=1,
        name=None,
    ):
        """Initializes the Interactions object."""
        self.window = window
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            name=name,
        )
        self.data = self.data.astype({"i": "int", "j": "int"})
        self.reset_mask()

    def mask_on_sequence(self, compliment_only=None, nts=None):
        """Mask interactions based on sequence.

        Parameters
        ----------
        compliment_only : bool, defaults to None
            If True, only keep interactions where i and j are complimentary
            nucleotides.
        nts : str, defaults to None
            If compliment_only is False, only keep interactions where i and j
            are in nts.

        Returns
        -------
        numpy array
            a boolean array of the same length as self.data
        """
        mask = []
        compliment = {"A": "UT", "U": "AG", "T": "AG", "G": "CU", "C": "G"}
        for _, i, j in self.data[["i", "j"]].itertuples():
            keep = []
            for w in range(self.window):
                i_nt = self.sequence[i + w - 1].upper()
                j_nt = self.sequence[j + self.window - w - 2].upper()
                if compliment_only:
                    keep.append(i_nt in compliment[j_nt])
                if nts is not None:
                    keep.append(i_nt in nts and j_nt in nts)
            mask.append(all(keep))
        self.update_mask(mask)
        return mask

    def mask_on_structure(
        self,
        structure,
        min_cd=None,
        max_cd=None,
        ss_only=False,
        ds_only=False,
        paired_only=False,
    ):
        """Masks interactions based on a secondary structure.

        Parameters
        ----------
        structure : rnavigate.data.SecondaryStructure
            The secondary structure to use for masking.
        min_cd : int, defaults to None
            The minimum contact distance to allow.
        max_cd : int, defaults to None
            The maximum contact distance to allow.
        ss_only : bool, defaults to False
            If True, only keep interactions between single-stranded nucleotides.
        ds_only : bool, defaults to False
            If True, only keep interactions between double-stranded nucleotides.
        paired_only : bool, defaults to False
            If True, only keep interactions that are paired in the structure.

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """
        if isinstance(structure, list):
            for each in structure:
                self.mask_on_structure(
                    each, min_cd, max_cd, ss_only, ds_only, paired_only
                )
                return
        alignment = data.SequenceAlignment(self, structure)
        mask = np.full(len(self.data), True)
        i_j_keep = ["i", "j", "mask"]
        for idx, i, j, keep in self.data[i_j_keep].itertuples():
            i = alignment.map_positions(i)
            j = alignment.map_positions(j)
            true_so_far = keep and (-1 not in [i, j])
            if paired_only and true_so_far:
                for w in range(self.window):
                    true_so_far = (
                        structure.pair_nts[i + w - 1] == j + self.window - w - 1
                    )
                    if not true_so_far:
                        break
            if (ss_only or ds_only) and true_so_far:
                percentage = 0
                for w in range(self.window):
                    percentage += int(structure.pair_nts[i - 1 + w] == 0)
                    percentage += int(structure.pair_nts[j - 1 + w] == 0)
                percentage /= self.window * 2
                if ss_only:
                    true_so_far = percentage > 0.501
                if ds_only:
                    true_so_far = percentage < 0.501
            if (min_cd is not None or max_cd is not None) and true_so_far:
                cd = []
                for iw in range(self.window):
                    for jw in range(self.window):
                        cd.append(structure.contact_distance(i + iw, j + jw))
                cd = min(cd)
                if min_cd is not None:
                    true_so_far = cd >= min_cd
                if max_cd is not None:
                    true_so_far = cd <= max_cd
            mask[idx] &= true_so_far
        self.update_mask(mask)
        return mask

    def mask_on_profile(self, profile, min_profile=None, max_profile=None):
        """Masks interactions based on per-nucleotide measurements.

        Parameters
        ----------
        profile : rnavigate.data.Profile
            The profile to use for masking.
        min_profile : float, defaults to None
            The minimum profile value to allow.
        max_profile : float, defaults to None
            The maximum profile value to allow.

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """
        mapping = data.SequenceAlignment(self, profile).mapping
        norm_prof = profile.data["Norm_profile"]
        mask = np.full(len(self.data), True)
        for idx, i, j in self.data[["i", "j"]].itertuples():
            index_i = mapping[i - 1]
            index_j = mapping[j - 1]
            keep_ij = True
            if (index_i != -1) and (index_j != -1):
                prof_i = np.nanmedian(norm_prof[index_i : index_i + self.window])
                prof_j = np.nanmedian(norm_prof[index_j : index_j + self.window])
                if min_profile is not None:
                    keep_ij &= (prof_i >= min_profile) | np.isnan(prof_i)
                    keep_ij &= (prof_j >= min_profile) | np.isnan(prof_j)
                if max_profile is not None:
                    keep_ij &= (prof_i <= max_profile) | np.isnan(prof_i)
                    keep_ij &= (prof_j <= max_profile) | np.isnan(prof_j)
            mask[idx] &= keep_ij
        self.update_mask(mask)
        return mask

    def mask_on_position(self, exclude=None, isolate=None):
        """Mask interactions based on their i and j positions.

        Parameters
        ----------
        exclude : list of int, defaults to None
            A list of positions to exclude.
        isolate : list of int, defaults to None
            A list of positions to isolate.

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """
        mask = np.full(len(self.data), True)
        for index, (i, j) in self.data[["i", "j"]].iterrows():
            if exclude is not None:
                if (i in exclude) or (j in exclude):
                    mask[index] = False
            if isolate is not None:
                if (i not in isolate) and (j not in isolate):
                    mask[index] = False
        self.update_mask(mask)
        return mask

    def mask_on_distance(self, max_dist=None, min_dist=None):
        """Mask interactions based on their distance in sequence space.

        Parameters
        ----------
        max_dist : int, defaults to None
            The maximum distance to allow. If None, no maximum distance is set.
        min_dist : int, defaults to None
            The minimum distance to allow. If None, no minimum distance is set.

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """
        primary_distances = self.data.eval("i - j").abs()
        mask = self.data["mask"]
        if min_dist is not None:
            mask &= primary_distances >= min_dist
        if max_dist is not None:
            mask &= primary_distances <= max_dist
        self.update_mask(mask)
        return mask

    def mask_on_values(self, **kwargs):
        """Mask interactions based on values in self.data.

        Parameters
        ----------
        kwargs : dict
            Each keyword should have the format "column_operator" where column
            is a valid column name of the dataframe and operator is one of:
                "ge": greater than or equal to
                "le": less than or equal to
                "gt": greater than
                "lt": less than
                "eq": equal to
                "ne": not equal to
            The values given to these keywords are then used in the comparison
            and False comparisons are filtered out. e.g.:
                self.mask_on_values(Statistic_ge=23) evaluates to:
                self.update_mask(self.data["Statistic"] >= 23)

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """
        mask = np.full(len(self.data), True)
        for key, value in kwargs.items():
            if key in self.data.keys():
                mask &= self.data[key] > value
            elif "_" in key:
                key2, comparison = key.rsplit("_", 1)
                operators = {"ge": ge, "le": le, "gt": gt, "lt": lt, "eq": eq, "ne": ne}
                if key2 in self.data.columns and comparison in operators:
                    operator = operators[comparison]
                    mask &= operator(self.data[key2], value)
                else:
                    print(f"{key}={value} is not a valid filter.")
            else:
                print(f"{key}={value} is not a valid filter.")
        self.update_mask(mask)
        return mask

    def reset_mask(self):
        """Resets the mask to all True (removes previous filters)"""
        self.data["mask"] = np.ones(len(self.data), dtype=bool)

    def copy(self, apply_filter=False):
        """Returns a copy of the interactions, optionally with masked rows removed.

        Parameters
        ----------
        apply_filter : bool, defaults to False
            If True, masked rows ("mask" == False) are dropped.

        Returns
        -------
        rnavigate.data.Interactions
            A copy of the interactions.
        """
        return self.get_aligned_data(
            alignment=self.null_alignment, apply_filter=apply_filter
        )

    def get_aligned_data(self, alignment, apply_filter=True):
        """Returns a copy mapped to a new sequence with masked rows removed.

        Parameters
        ----------
        alignment : rnavigate.data.SequenceAlignment
            The alignment to use for mapping the interactions.
        apply_filter : bool, defaults to True
            If True, masked rows ("mask" == False) are dropped.

        Returns
        -------
        rnavigate.data.Interactions
            Interactions mapped to a new sequence.
        """
        if apply_filter:
            dataframe = self.data[self.data["mask"]].reset_index(drop=True)
        else:
            dataframe = self.data
        new_data = alignment.map_dataframe(
            dataframe=dataframe, position_columns=["i", "j"]
        )
        return self.__class__(
            input_data=new_data,
            sequence=alignment.target_sequence,
            metric=self._metric,
            metric_defaults=self.metric_defaults,
            window=self.window,
            name=self.name,
        )

    def update_mask(self, mask):
        """Updates the mask by ANDing the current mask with the given mask."""
        self.data["mask"] = self.data["mask"] & mask

    def count_filter(self, **kwargs):
        """Counts the number of interactions that pass the given filters."""
        mask = self.filter(**kwargs)
        return sum(mask)

    def filter(
        self,
        prefiltered=False,
        reset_filter=True,
        # mask on structure
        structure=None,
        min_cd=None,
        max_cd=None,
        paired_only=False,
        ss_only=False,
        ds_only=False,
        # mask on profile
        profile=None,
        min_profile=None,
        max_profile=None,
        # mask on sequence
        compliments_only=False,
        nts=None,
        # mask on position
        max_distance=None,
        min_distance=None,
        exclude_nts=None,
        isolate_nts=None,
        # others
        resolve_conflicts=None,
        **kwargs,
    ):
        """Convenience function that applies the above filters simultaneously.

        Parameters
        ----------
        prefiltered : bool, defaults to False
            If True, the mask is not updated.
        reset_filter : bool, defaults to True
            If True, the mask is reset before applying filters.
        structure : rnavigate.data.SecondaryStructure, defaults to None
            The structure to use for filtering.
        min_cd : int, defaults to None
            The minimum contact distance to allow.
        max_cd : int, defaults to None
            The maximum contact distance to allow.
        paired_only : bool, defaults to False
            If True, only keep interactions that are paired in the structure.
        ss_only : bool, defaults to False
            If True, only keep interactions between single-stranded nucleotides.
        ds_only : bool, defaults to False
            If True, only keep interactions between double-stranded nucleotides.
        profile : rnavigate.data.Profile, defaults to None
            The profile to use for masking.
        min_profile : float, defaults to None
            The minimum profile value to allow.
        max_profile : float, defaults to None
            The maximum profile value to allow.
        compliments_only : bool, defaults to False
            If True, only keep interactions where i and j are complimentary
            nucleotides.
        nts : str, defaults to None
            If compliment_only is False, only keep interactions where i and j
            are in nts.
        max_distance : int, defaults to None
            The maximum distance to allow. If None, no maximum distance is set.
        min_distance : int, defaults to None
            The minimum distance to allow. If None, no minimum distance is set.
        exclude_nts : list of int, defaults to None
            A list of positions to exclude.
        isolate_nts : list of int, defaults to None
            A list of positions to isolate.
        resolve_conflicts : str, defaults to None
            If not None, conflicting windows are resolved using the Maximal
            Weighted Independent Set. The weights are taken from the metric
            value. The graph is first broken into components to speed up the
            identification of the MWIS. Then the mask is updated to only
            include the MWIS.
        **kwargs : dict
            Each keyword should have the format "column_operator" where column
            is a valid column name of the dataframe and operator is one of:
                "ge": greater than or equal to
                "le": less than or equal to
                "gt": greater than
                "lt": less than
                "eq": equal to
                "ne": not equal to
            The values given to these keywords are then used in the comparison
            and False comparisons are filtered out. e.g.:
                self.mask_on_values(Statistic_ge=23) evaluates to:
                self.update_mask(self.data["Statistic"] >= 23)

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """

        def filters_are_on(*filters):
            return any(f not in [None, False] for f in filters)

        if prefiltered:
            return
        if reset_filter:
            self.reset_mask()
        mask = np.full(len(self.data), True)
        if filters_are_on(exclude_nts, isolate_nts):
            mask &= self.mask_on_position(exclude=exclude_nts, isolate=isolate_nts)
        if filters_are_on(max_distance, min_distance):
            mask &= self.mask_on_on_distance(
                max_dist=max_distance, min_dist=min_distance
            )
        if filters_are_on(min_profile, max_profile):
            mask &= self.mask_on_profile(
                profile=profile, min_profile=min_profile, max_profile=max_profile
            )
        if filters_are_on(min_cd, max_cd, ss_only, ds_only, paired_only):
            mask &= self.mask_on_structure(
                structure=structure,
                min_cd=min_cd,
                max_cd=max_cd,
                ss_only=ss_only,
                ds_only=ds_only,
                paired_only=paired_only,
            )
        if filters_are_on(compliments_only, nts):
            mask &= self.mask_on_sequence(compliment_only=compliments_only, nts=nts)
        kwargs, data_specific_mask = self.data_specific_filter(**kwargs)
        mask &= data_specific_mask
        mask &= self.mask_on_values(**kwargs)
        if resolve_conflicts is not None:
            mask &= self.resolve_conflicts(metric=resolve_conflicts)
        return mask

    def data_specific_filter(self, **kwargs):
        """Does nothing for the base Interactions class, can be overwritten in
        subclasses.

        Returns:
            dict: dictionary of keyword argument pairs
        """
        return kwargs, np.full(len(self.data), True)

    def get_ij_colors(self):
        """Gets i, j, and colors lists for plotting interactions.

        i and j are the 5' and 3' ends of each interaction, and colors is the color
        to use for each interaction. Values of self.data[self.metric] are normalized
        to 0 to 1, which correspond to self.min_max values. These are then mapped to
        a color using self.cmap.

        Returns
        -------
        i : list
            5' ends of each interaction
        j : list
            3' ends of each interaction
        colors : list
            colors to use for each interaction
        """
        if len(self.data) == 0:
            return [], [], []

        dataframe = self.get_sorted_data()
        i_list, j_list, colors = [], [], []
        for _, i, j, datum in dataframe[["i", "j", self.metric]].itertuples():
            for w in range(self.window):
                i_list.append(i + w)
                j_list.append(j + self.window - 1 - w)
                colors.append(datum)
        colors = self.cmap.values_to_hexcolors(colors, 0.6)
        return i_list, j_list, colors

    def get_sorted_data(self):
        """Returns a copy of the data sorted by self.metric.

        Returns
        -------
        pandas.DataFrame
            a copy of the data sorted by self.metric
        """
        ascending = self.metric not in ["Distance"]  # high distance is bad
        dataframe = self.data[["i", "j", self.metric]].copy()
        dataframe.sort_values(
            by=self.metric, ascending=ascending, inplace=True, na_position="first"
        )
        return dataframe

    def print_new_file(self, outfile=None):
        """Create a new file with mapped and filtered interactions.

        Parameters
        ----------
        outfile : str, defaults to None
            path to an output file. If None, file string is printed to console.
        """
        data = self.data.copy()
        data = data[data["mask"]]
        if "Sign" in data.columns:
            data.rename({"Sign": "+/-"}, inplace=True)
        csv = data.to_csv(
            columns=self.columns, sep="\t", index=False, line_terminator="\n"
        )
        if outfile is not None:
            with open(outfile, "w") as out:
                out.write(self.header)
                out.write(csv)
        else:
            print(self.header, csv)

    def set_3d_distances(self, pdb, atom):
        """Calculates the distance between atoms in i and j in the PDB structure.

        Parameters
        ----------
        pdb : rnavigate.pdb.PDB
            PDB object to use for calculating distances
        atom : str
            atom id to use for calculating distances
        """
        mapping = data.SequenceAlignment(self, pdb).mapping
        distance_matrix = pdb.get_distance_matrix(atom=atom)
        i = self.data["i"].values
        j = self.data["j"].values
        distances = np.zeros(len(i))
        pairs = self.window**2
        for iw in range(self.window):
            for jw in range(self.window):
                io = mapping[i + iw - 1]
                jo = mapping[j + jw - 1]
                distances += distance_matrix[io, jo] / pairs
        self.data["Distance"] = distances

    def resolve_conflicts(self, metric=None):
        """Uses an experimental method to resolve conflicts.

        Resolves conflicting windows using the Maximal Weighted Independent
        Set. The weights are taken from the metric value. The graph is first
        broken into components to speed up the identification of the MWIS. Then
        the mask is updated to only include the MWIS. This method is computationally
        expensive for large or dense datasets.

        Parameters
        ----------
        metric : str, defaults to None
            The metric to use for weighting the graph. If None, self.metric is used.

        Returns
        -------
        mask : numpy array
            a boolean array of the same length as self.data
        """

        def get_components(graph):
            """Using DFS, returns a list of component subgraphs of graph."""

            def dfs(key, value, component):
                """Depth first seach to identify a component."""
                visited[key] = True
                component[key] = value
                for v in value:
                    if not visited[v]:
                        component = dfs(v, graph[v], component)
                return dict(component)

            components = []
            visited = {key: False for key in graph.keys()}
            for key, value in graph.items():
                if not visited[key]:
                    components.append(dfs(key, value, {}))
            return components

        def graphSets(graph, node_weights):
            """Finds the Maximal Weighted Independent Set of a graph."""
            # Base Case - Given Graph has no nodes
            if len(graph) == 0:
                return []
            # Select a vertex from the graph
            current_vertex = list(graph.keys())[0]
            # Create a copy of the graph and remove the current vertex
            new_graph = dict(graph)
            del new_graph[current_vertex]
            # Case 1 - excluding current Vertex
            case_1 = graphSets(new_graph, node_weights)
            # Case 2 - including current Vertex, and excluding neighbors
            for neighbor in graph[current_vertex]:
                if neighbor in new_graph:
                    del new_graph[neighbor]
            case_2 = [current_vertex] + graphSets(new_graph, node_weights)
            # Our final result is the one which has highest weight, return it
            weight_1 = sum(node_weights[node] for node in case_1)
            weight_2 = sum(node_weights[node] for node in case_2)
            if weight_1 > weight_2:
                return case_1
            return case_2

        # Building the graph ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        if metric is None:
            metric = self.metric
        # a list of indices relating back to the original mask
        indices = np.where(self.data["mask"])[0]
        # the correlation data to be considered
        df = self.data.loc[indices].copy()
        window = self.window
        # node_weight[i from indices] = metric value for that correlation
        node_weights = {}
        # graph[i from indices] = list of indices of conflicting correlations
        graph = {}
        # for each considered correlation, record conflicts and node weights
        for v1, i, j, g in df[["i", "j", metric]].itertuples():
            if v1 not in graph:
                graph[v1] = []
            node_weights[v1] = g  # recording the node weight
            # conflicts = indices of all overlapping, but not parallel corrs
            conflicts = np.where(
                (
                    (abs(df["i"] - i) < window)
                    | (abs(df["j"] - i) < window)
                    | (abs(df["i"] - j) < window)
                    | (abs(df["j"] - j) < window)
                )
                & ((df["j"] - j) != (i - df["i"]))
            )[0]
            # adding conflicts to graph dictionary
            for conflict in conflicts:
                v2 = indices[conflict]
                if v2 > v1:  # prevents self-loop and parallel edges
                    if v2 not in graph:
                        graph[v2] = []
                    graph[v1].append(v2)
                    graph[v2].append(v1)
        # Finding the maximum set ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Split the graph into components. Get the max set for each component.
        # The graph's max set is the same as the union of these max sets.
        max_set = []
        for component in get_components(graph):
            max_set += graphSets(component, node_weights)
        # set these indices to true and update the original mask
        new_mask = np.zeros(len(self.data), dtype=bool)
        new_mask[max_set] = True
        self.update_mask(new_mask)
        return new_mask


class SHAPEJuMP(Interactions):
    """A class for storing and manipulating SHAPEJuMP data.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing SHAPEJuMP data.
        If dataframe, the dataframe containing SHAPEJuMP data. The dataframe
        must contain columns "i", "j", "Metric" (JuMP rate) and "Percentile"
        (percentile ranking). Dataframe may also include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the SHAPEJuMP data.
    metric : string, defaults to "Percentile"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap)
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict
        kwargs passed to pandas.read_table() when reading input_data.
    window : int
        The window size used to generate the SHAPEJuMP data.
    name : str
        A name for the interactions object.

    Attributes
    ----------
    data : pandas.DataFrame
        The SHAPEJuMP data.
    """

    def __init__(
        self,
        input_data,
        sequence=None,
        metric="Percentile",
        metric_defaults=None,
        read_table_kw=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from SHAPEJuMP data"""
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Percentile": {
                "metric_column": "Percentile",
                "cmap": "YlGnBu",
                "normalization": "min_max",
                "values": [0.98, 1.0],
                "title": "SHAPE-JuMP: percentile",
                "extend": "min",
            },
            "Metric": {
                "metric_column": "Metric",
                "cmap": "YlGnBu",
                "normalization": "min_max",
                "values": [0, 0.001],
                "extend": "max",
                "title": "SHAPE-JuMP: rate",
            },
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            window=window,
            name=name,
        )

    def read_file(self, input_data, read_table_kw=None):
        """Parses a deletions.txt file and stores it as a dataframe.

        Also calculates a "Percentile" column.

        Parameters
        ----------
        input_data : str
            path to deletions.txt file
        read_table_kw : dict, defaults to {}
            kwargs passed to pandas.read_table().

        Returns
        -------
        pandas.DataFrame
            the SHAPEJuMP data
        """
        column_names = ["Gene", "i", "j", "Metric"]
        data = pd.read_table(input_data, names=column_names, header=0, **read_table_kw)
        data["Percentile"] = data["Metric"].rank(method="max", pct=True)
        return data


class RINGMaP(Interactions):
    """A class for storing and manipulating RINGMaP data.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing RINGMaP data.
        If dataframe, the dataframe containing RINGMaP data. The dataframe
        must contain columns "i", "j", "Statistic", and "Zij". Dataframe may
        also include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the RINGMaP data.
    metric : string, defaults to "Statistic"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap)
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the RINGMaP data. If an input file is
        provided, this value is overwritten by the value in the header.
    name : str, optional
        A name for the interactions object.

    Attributes
    ----------
    data : pandas.DataFrame
        The RINGMaP data.
    """

    def __init__(
        self,
        input_data,
        sequence=None,
        metric="Statistic",
        metric_defaults=None,
        read_table_kw=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from RINGMaP data"""
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Statistic": {
                "metric_column": "Statistic",
                "cmap": "bwr",
                "normalization": "min_max",
                "values": [-100, 100],
                "title": "RING-MaP: Gapc",
                "extend": "both",
            },
            "Zij": {
                "metric_column": "Zij",
                "cmap": "bwr",
                "normalization": "min_max",
                "values": [-8, 8],
                "title": "RING-MaP: Zij",
                "extend": "both",
            },
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            window=window,
            name=name,
        )

    def read_file(self, filepath, read_table_kw=None):
        """Parses a RINGMaP correlations file and stores data as a dataframe.

        Also sets self.window (usually 1, from header).

        Parameters
        ----------
        filepath : str
            path to correlations file.
        read_table_kw : dict, defaults to {}
            kwargs passed to pandas.read_table().

        Returns
        -------
        pandas.DataFrame
            the RINGMaP data
        """
        with open(filepath, "r") as file:
            self.header = file.readline()
        split_header = self.header.split("\t")
        window = split_header[1].split("=")[1]
        self.window = int(window)
        dataframe = pd.read_table(filepath, header=1, **read_table_kw)
        dataframe.rename(columns={"+/-": "Sign"}, inplace=True)
        return dataframe

    def data_specific_filter(self, positive_only=False, negative_only=False, **kwargs):
        """Adds filters for "Sign" column to parent filter() function

        Parameters
        ----------
        positive_only : bool, defaults to False
            If True, only keep positive correlations.
        negative_only : bool, defaults to False
            If True, only keep negative correlations.

        Returns
        -------
        kwargs : dict
            any additional keyword-argument pairs are returned
        mask : numpy array
            a boolean array of the same length as self.data
        """
        mask = np.full(len(self.data), True)
        if positive_only:
            mask &= self.data["Sign"] == 1
        if negative_only:
            mask &= self.data["Sign"] == -1
        self.update_mask(mask)
        return kwargs, mask

    def get_sorted_data(self):
        """Sorts on the product of self.metric and "Sign" columns.

        Except when self.metric is "Distance".

        Returns
        -------
        pandas.DataFrame
            a copy of the data sorted by (self.metric * "Sign") columns
        """
        if self.metric == "Distance":
            return super().get_sorted_data()
        columns = ["i", "j", self.metric]
        dataframe = self.data[columns + ["Sign"]].copy()
        dataframe.eval(f"{self.metric} = {self.metric} * Sign", inplace=True)
        dataframe.sort_values(by=self.metric, ascending=True, inplace=True)
        return dataframe[columns]


class PAIRMaP(RINGMaP):
    """A class for storing and manipulating PAIRMaP data.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing PAIRMaP data.
        If dataframe, the dataframe containing PAIRMaP data. The dataframe
        must contain columns "i", "j", "Statistic", and "Class". Dataframe may
        also include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the PAIRMaP data.
    metric : string, defaults to "Class"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap)
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the PAIRMaP data. If an input file is
        provided, this value is overwritten by the value in the header.
    name : str, optional
        A name for the interactions object.

    Attributes
    ----------
    data : pandas.DataFrame
        The PAIRMaP data.
    """

    def __init__(
        self,
        input_data,
        sequence=None,
        metric="Class",
        metric_defaults=None,
        read_table_kw=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from PAIRMaP data"""
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Class": {
                "metric_column": "Class",
                "cmap": mpc.ListedColormap(
                    [[0.7, 0.7, 0.7], [0.0, 0.0, 0.95], [0.12, 0.76, 1.0]]
                ),
                "normalization": "none",
                "values": None,
                "title": "PAIR-MaP",
                "extend": "both",
                "ticks": [0, 1, 2],
                "tick_labels": ["Complimentary", "Primary", "Secondary"],
            }
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            window=window,
            name=name,
        )

    def read_file(self, filepath, read_table_kw=None):
        """Parses a pairmap.txt file and stores data as a dataframe

        Sets self.window (usually 3), from header.

        Parameters
        ----------
        filepath : str
            path to pairmap.txt file
        read_table_kw : dict, defaults to None
            This argument is ignored.
        """
        with open(filepath, "r") as file:
            self.header = file.readline()
        self.window = int(self.header.split("\t")[1].split("=")[1])
        dataframe = pd.read_table(filepath, header=1)
        dataframe.rename(columns={"Sig.": "Statistic"}, inplace=True)
        dataframe["Sign"] = 1
        return dataframe

    def data_specific_filter(self, all_pairs=False, **kwargs):
        """Used by Interactions.filter(). By default, non-primary and
        -secondary pairs are removed. all_pairs=True changes this behavior.

        Parameters
        ----------
        all_pairs : bool, defaults to False
            whether to include all PAIRs.

        Returns
        -------
        kwargs : dict
            any additional keyword-argument pairs are returned
        mask : numpy array
            a boolean array of the same length as self.data
        """
        mask = np.full(len(self.data), True)
        if not all_pairs:
            mask &= self.data["Class"] != 0
        self.update_mask(mask)
        return kwargs, mask

    def get_sorted_data(self):
        """Same as parent function, unless metric is set to "Class", in which
        case ij pairs are returned in a different order.

        Returns
        -------
        pandas.DataFrame
            a copy of the data sorted by self.metric
        """
        if self.metric != "Class":
            return super().get_sorted_data()
        metric = self.metric
        columns = ["i", "j", metric]
        dataframe = self.data[columns].copy()
        # Class data have a weird order
        dataframe = pd.concat(
            [
                dataframe[dataframe[metric] == 0],
                dataframe[dataframe[metric] == 2],
                dataframe[dataframe[metric] == 1],
            ]
        )
        return dataframe


class PairingProbability(Interactions):
    """A class for storing and manipulating pairing probability data.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing pairing probability data.
        If dataframe, the dataframe containing pairing probability data. The
        dataframe must contain columns "i", "j", "Probability", and "log10p".
        Dataframe may also include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the pairing probability data.
    metric : string, defaults to "Probability"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the pairing probability data.
    name : str, optional
        A name for the PairingProbability object.

    Attributes
    ----------
    data : pandas.DataFrame
        The pairing probability data.
    """

    def __init__(
        self,
        input_data,
        sequence=None,
        metric="Probability",
        metric_defaults=None,
        read_table_kw=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from pairing probability data"""
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "Probability": {
                "metric_column": "Probability",
                "cmap": sns.cubehelix_palette(
                    10, 0.7, 0.9, 1.5, 2.5, 1, 0.4, False, True
                ),
                "normalization": "min_max",
                "values": [0, 1],
                "title": "Pairing probability",
                "extend": "neither",
            },
            "Probability_old": {
                "metric_column": "Probability",
                "cmap": mpc.ListedColormap(
                    [(150, 150, 150), (255, 204, 0), (72, 143, 205), (81, 184, 72)]
                ),
                "normalization": "bins",
                "values": [0.1, 0.3, 0.8],
                "extend": "neither",
                "title": "Pairing probability",
            },
            "Probability_continuous": {
                "metric_column": "Probability",
                "cmap": "plasma_r",
                "normalization": "min_max",
                "values": [0.0, 1.0],
                "extend": "neither",
                "title": "Pairing probability",
            },
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            window=window,
            name=name,
        )

    def read_file(self, filepath, read_table_kw=None):
        """Parses a pairing probability file and stores data as a dataframe.

        Parameters
        ----------
        filepath : str
            path to pairing probability file
        read_table_kw : dict, defaults to None
            This argument is ignored.

        Returns
        -------
        pandas.DataFrame
            the pairing probability data
        """
        with open(filepath, "r") as file:
            self.header = file.readline()
        self.window = 1
        dataframe = pd.read_table(filepath, header=1, names=["i", "j", "log10p"])
        dataframe["Probability"] = 10 ** (-dataframe["log10p"])
        return dataframe

    def get_entropy_profile(self, print_out=False, save_file=None):
        """Calculates per-nucleotide Shannon entropy from pairing probabilities.

        Parameters
        ----------
        print_out : bool, defaults to False
            If True, entropy values are printed to console.
        save_file : str, defaults to None
            If not None, entropy values are saved to this file.

        Returns
        -------
        rnavigate.data.Profile
            a Profile object containing the entropy data
        """
        self.data.eval("nlogn = log10p * 10 ** ( - log10p )", inplace=True)
        entropy = np.zeros(self.length)
        args = {"labels": np.arange(self.length) + 1, "fill_value": 0}
        i_sum = self.data[["i", "nlogn"]].groupby("i").sum().reindex(**args)
        j_sum = self.data[["j", "nlogn"]].groupby("j").sum().reindex(**args)
        ij_sum = i_sum + j_sum
        entropy_df = pd.DataFrame(
            {
                "Sequence": list(self.sequence),
                "Nucleotide": np.arange(self.length) + 1,
                "Entropy": ij_sum["nlogn"],
            }
        )
        entropy_df = entropy_df.set_index("Nucleotide")
        # catch rounding errors:
        entropy_df.loc[entropy_df["Entropy"] < 0, "Entropy"] = 0
        if print_out:
            print(*[f"{i+1} {s}" for i, s in enumerate(entropy)], sep="\n")
        if save_file is not None:
            with open(save_file) as outf:
                for i, s in enumerate(entropy):
                    outf.write(f"{i+1}\t{s}\n")
        return data.Profile(
            input_data=entropy_df,
            metric="Entropy",
            metric_defaults={
                "Entropy": {
                    "metric_column": "Entropy",
                    "error_column": None,
                    "color_column": None,
                    "cmap": "rainbow",
                    "normalization": "norm",
                    "values": [0.0, 0.2],
                    "extend": "right",
                    "title": "Shannon entropy",
                    "alpha": 1,
                }
            },
        )

    def data_specific_filter(self, **kwargs):
        """By default, interactions with probabilities less than 0.03 are removed.

        Returns
        -------
        kwargs : dict
            any additional keyword-argument pairs are returned
        mask : numpy array
            a boolean array of the same length as self.data
        """
        mask = self.data["Probability"] >= 0.03
        self.update_mask(mask)
        return kwargs, mask


class AllPossible(Interactions):
    """A class for storing and manipulating all possible interactions.

    Parameters
    ----------
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the pairing probability data.
    metric : string, defaults to "Probability"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the pairing probability data.
    name : str, optional
        A name for the PairingProbability object.

    Attributes
    ----------
    data : pandas.DataFrame
        The pairing probability data.
    """

    def __init__(
        self,
        sequence,
        metric="data",
        input_data=None,
        metric_defaults=None,
        read_table_kw=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from pairing probability data"""
        if isinstance(sequence, data.Sequence):
            sequence = sequence.sequence
        if metric_defaults is None:
            metric_defaults = {}
        metric_defaults = {
            "data": {
                "metric_column": "data",
                "cmap": "magenta",
                "normalization": "none",
                "values": None,
                "title": "Hypthetical pairs",
                "extend": "neither",
            }
        } | metric_defaults
        if input_data is not None:
            dataframe = input_data
        else:
            dataframe = {"i": [], "j": [], "data": []}
            for i in range(1, len(sequence)):
                for j in range(i + 1, len(sequence) + 1):
                    dataframe["i"].append(i)
                    dataframe["j"].append(j)
                    dataframe["data"].append(0)
            dataframe = pd.DataFrame(dataframe)
        super().__init__(
            input_data=dataframe,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            read_table_kw=read_table_kw,
            window=window,
            name=name,
        )


class StructureAsInteractions(Interactions):
    """A class for storing and manipulating structure data.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing structure data.
        If dataframe, the dataframe containing structure data. The dataframe
        must contain columns "i", "j", and "Structure". Dataframe may also
        include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the structure data.
    metric : string, defaults to "Structure"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the structure data.
    name : str, optional
        A name for the StructureAsInteractions object.

    Attributes
    ----------
    data : pandas.DataFrame
        The structure data.
    """

    def __init__(
        self,
        input_data,
        sequence,
        metric=None,
        metric_defaults=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from structure data"""
        if metric_defaults is None:
            metric_defaults = {}
        if metric is None:
            metric = "Structure"
        if isinstance(input_data, pd.DataFrame):
            pass
        elif isinstance(input_data, data.SecondaryStructure):
            input_data = input_data.get_interactions_df()
        metric_defaults = {
            "Structure": {
                "metric_column": "Structure",
                "cmap": "grey",
                "normalization": "none",
                "ticks": [],
                "title": "Base-pairs",
                "extend": "neither",
            }
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            window=window,
            name=name,
        )


class StructureCompareTwo(Interactions):
    """A class for storing and manipulating a comparison of two structures.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing structure data.
        If dataframe, the dataframe containing structure data. The dataframe
        must contain columns "i", "j", and "Structure". Dataframe may also
        include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the structure data.
    metric : string, defaults to "Structure"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the structure data.
    name : str, optional
        A name for the StructureAsInteractions object.

    Attributes
    ----------
    data : pandas.DataFrame
        The structure data.
    """

    def __init__(
        self,
        input_data,
        sequence,
        metric=None,
        metric_defaults=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from structure data"""
        if metric is None:
            metric = "Which_structure"
        if metric_defaults is None:
            metric_defaults = {}
        if isinstance(input_data, pd.DataFrame):
            pass
        elif isinstance(input_data, list) and len(input_data) == 2:
            ss1, ss2 = [ss.get_interactions_df() for ss in input_data]
            input_data = ss1.merge(
                ss2,
                how="outer",
                on=["i", "j"],
                indicator="Which_structure",
                suffixes=["_left", "_right"],
            )
            categories = {"both": 0, "left_only": 1, "right_only": 2}
            input_data["Which_structure"] = [
                categories[c] for c in input_data["Which_structure"]
            ]
            input_data["Which_structure"].astype(int)
            input_data = input_data.reset_index(drop=True)
        metric_defaults = {
            "Which_structure": {
                "metric_column": "Which_structure",
                "cmap": [
                    # shared
                    (150 / 255.0, 150 / 255.0, 150 / 255.0),
                    # left
                    (38 / 255.0, 202 / 255.0, 145 / 255.0),
                    # right
                    (153 / 255.0, 0.0, 1.0),
                ],
                "normalization": "none",
                "values": [],
                "title": "Base-pairs by structure",
                "extend": "neither",
                "ticks": [0, 1, 2],
                "tick_labels": [
                    "common\npairs",
                    "first\nstructure",
                    "second\nstructure",
                ],
            }
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            window=window,
            name=name,
        )


class StructureCompareMany(Interactions):
    """A class for storing and manipulating a comparison of many structures.

    Parameters
    ----------
    input_data : string or pandas.DataFrame
        If string, a path to a file containing structure data.
        If dataframe, the dataframe containing structure data. The dataframe
        must contain columns "i", "j", and "Structure". Dataframe may also
        include other columns.
    sequence : string or rnavigate.data.Sequence
        The sequence string corresponding to the structure data.
    metric : string, defaults to "Structure"
        The column name to use for visualization.
    metric_defaults : dict
        Keys are metric names and values are dictionaries of metric-specific defaults.
        These defaults include:
            "metric_column" : string
                the column name to use for visualization
            "cmap" : string or matplotlib.colors.Colormap
                the colormap to use for visualization
            "normalization" : "min_max", "0_1", "none", or "bins"
                The type of normalization to use when mapping values to colors
            "values" : list of float
                The values to used with normalization of the data
            "title" : string
                the title to use for colorbars
            "extend" : "min", "max", "both", or "neither"
                Which ends to extend when drawing the colorbar.
            "tick_labels" : list of string
    read_table_kw : dict, optional
        kwargs passed to pandas.read_table() when reading input_data.
    window : int, defaults to 1
        The window size used to generate the structure data.
    name : str, optional
        A name for the StructureAsInteractions object.

    Attributes
    ----------
    data : pandas.DataFrame
        The structure data.
    """

    def __init__(
        self,
        input_data,
        sequence,
        metric=None,
        metric_defaults=None,
        window=1,
        name=None,
    ):
        """Constructs an Interactions object from structure data"""
        if metric is None:
            metric = "Num_structures"
        if metric_defaults is None:
            metric_defaults = {}
        if isinstance(input_data, pd.DataFrame):
            pass
        elif isinstance(input_data, list) and len(input_data) >= 2:
            input_data = [ss.get_interactions_df() for ss in input_data]
            other_structures = input_data[1:]
            input_data = input_data[0]
            columns = ["Structure_1"]
            input_data = input_data.rename(columns={"Structure": "Structure_1"})
            for i, structure in enumerate(other_structures):
                col = f"Structure_{i+2}"
                columns.append(col)
                structure = structure.rename(columns={"Structure": col})
                input_data = input_data.merge(structure, how="outer", on=["i", "j"])
            input_data[columns] = input_data[columns].fillna(0)
            input_data["Num_structures"] = (
                input_data[columns].sum(axis=1, numeric_only=True) - 1
            )  # to index at 0 for coloring
            input_data = input_data.reset_index(drop=True)
        total_structures = input_data["Num_structures"].max() + 1
        metric_defaults = {
            "Num_structures": {
                "metric_column": "Num_structures",
                "cmap": sns.color_palette("rainbow_r", total_structures),
                "normalization": "none",
                "values": [],
                "title": "Base-pairs by structure",
                "extend": "neither",
                "ticks": [i for i in range(total_structures)],
                "tick_labels": [i + 1 for i in range(total_structures)],
            }
        } | metric_defaults
        super().__init__(
            input_data=input_data,
            sequence=sequence,
            metric=metric,
            metric_defaults=metric_defaults,
            window=window,
            name=name,
        )



import pandas as pd
from rnavigate import data, transcriptomics
import requests
from pathlib import Path


data_path = Path(__file__).parent.parent.parent
data_path /= "reference_data/eCLIP_downloads"


def download_eclip_peaks(assembly="GRCh38", outpath=data_path):
    """download eCLIP bed files from ENCODE database

    Args:
        assembly (string, optional): reference genome ("h19" or "GRCh38")
            Defaults to "GRCh38"
        outpath (string, optional): output directory path
            Defaults to "eCLIP_downloads".
    """
    if Path.is_dir(outpath):
        print(
            f"{outpath} folder already exists. "
            "Files will be downloaded the same folder."
        )
    else:
        Path.mkdir(outpath, parents=True)
        print(f"Created {outpath} folder.")

    print("Searching ENCODE database for eCLIP experiments")
    accession = []
    bed_accession = []

    # Force return from the server in JSON format
    headers = {"accept": "application/json"}
    # GET accession number for experiments
    url_bed = (
        "https://www.encodeproject.org/search/?"
        "type=Experiment&"
        "assay_title=eCLIP&"
        "frame=object&"
        "limit=all&"
        "files.file_type=bed+narrowPeak"
    )
    response = requests.get(url_bed, headers=headers, timeout=21.1)
    # Load json response as a Python dictionary and append accession to list
    experiment_dict = response.json()
    for entry in experiment_dict["@graph"]:
        accession.append(entry["accession"])

    # GET accession number for IDR bed files, store in a dict.
    # Key is experiment acc and values are bed files acc for that experiment.
    for exp in accession:
        url2 = (
            "https://www.encodeproject.org/search/?"
            "type=File&"
            f"dataset=/experiments/{exp}/&"
            "file_format=bed&"
            "format=json&"
            "frame=object&"
            "limit=all"
        )
        response2 = requests.get(url2, headers=headers, timeout=21.1)
        bed_dict = response2.json()
        for i in bed_dict["@graph"]:
            # choose correct assembly and choose replicable peaks
            # if want a specific replicate then change
            # i["biological_replicates"] == "rep_1" or "rep_2"
            if (len(i["biological_replicates"]) == 2) and (i["assembly"] == assembly):
                bed_accession.append(i["accession"])

    # download all IDR bed files in accession
    for acc in bed_accession:
        download_url = f"https://www.encodeproject.org/files/{acc}/@@download/{acc}.bed.gz"  # noqa: E501 pylint: disable=line-too-long
        bedgzfile = requests.get(download_url, timeout=21.1)
        if outpath:
            outfile = outpath + f"/{acc}.bed.gz"
        else:
            outfile = f"/{acc}.bed.gz"
        open(outfile, "wb").write(bedgzfile.content)
    print(f"Download finished. All files are stored in {outpath} dir.")

    # create a table to look up eCLIP files using target and cell type.
    files = Path.iterdir(outpath)
    codes = {"accession": [], "target": [], "cell_line": []}
    for file in files:
        if not file.endswith("bed.gz"):
            continue
        df = pd.read_table(file, compression="gzip", header=None)
        label = df.iloc[0, 3]
        label = label.split("_")
        if len(label) == 1:
            label = ["", "", ""]
        elif len(label) == 4:
            label = [label[0], f"{label[1]}_{label[2]}", label[3]]
        codes["accession"].append(file.rstrip(".bed.gz"))
        codes["target"].append(label[0])
        codes["cell_line"].append(label[1])
    codes = pd.DataFrame(codes)
    k562 = codes[codes["cell_line"] == "K562", ["target", "accession"]]
    k562.columns = ["target", "K562"]
    hepg2 = codes[codes["cell_line"] == "HepG2", ["target", "accession"]]
    hepg2.columns = ["target", "HepG2"]
    df = pd.merge(k562, hepg2, how="outer", on="target")
    df["target", "K562", "HepG2"].to_csv(
        outpath / "eclip_codes.txt", index=False, sep="\t"
    )


class eCLIPDatabase:  # pylint disable=invalid-name
    def __init__(self):
        self.path = data_path
        self.eclip_codes = pd.read_table(self.path / "eclip_codes.txt")
        self.eclip_data = self.get_eclip_data()

    def get_eclip_data(self):
        eclip_data = {"HepG2": {}, "K562": {}}
        hepg2_rows = ~self.eclip_codes["HepG2"].isnull()
        k562_rows = ~self.eclip_codes["K562"].isnull()
        for _, row in self.eclip_codes[hepg2_rows].iterrows():
            eclip_data["HepG2"][row["target"]] = transcriptomics.NarrowPeak(
                str(self.path / f"{row['HepG2']}.bed.gz")
            )
        for _, row in self.eclip_codes[k562_rows].iterrows():
            eclip_data["K562"][row["target"]] = transcriptomics.NarrowPeak(
                str(self.path / f"{row['K562']}.bed.gz")
            )
        return eclip_data

    def get_eclip_density(self, transcript, cell_line, targets=None):
        eclip_data = self.eclip_data[cell_line]
        if targets is not None:
            eclip_data = {target: eclip_data[target] for target in targets}
        eclip_density = data.Profile(
            input_data=transcript.coordinate_df.eval("eCLIP_density = 0"),
            metric="eCLIP_density",
        )
        for value in eclip_data.values():
            values = value.get_profile(transcript).data.eval("Value != 0.0")
            eclip_density.data["eCLIP_density"] += values
        return eclip_density

    def get_annotation(self, transcript, cell_line, target):
        self.eclip_data[cell_line][target].get_annotation(transcript)

    def get_profile(self, transcript, cell_line, target):
        self.eclip_data[cell_line][target].get_profile(transcript)



from rnavigate.data.data import Sequence, Data, normalize_sequence
from rnavigate.data.secondary_structure import (
    SecondaryStructure,
    SequenceCircle,
    StructureCoordinates,
)
from rnavigate.data.alignments import (
    AlignmentChain,
    lookup_alignment,
    set_alignment,
    set_multiple_sequence_alignment,
    SequenceAlignment,
    StructureAlignment,
)
from rnavigate.data.colors import ScalarMappable
from rnavigate.data.interactions import (
    Interactions,
    SHAPEJuMP,
    RINGMaP,
    PAIRMaP,
    PairingProbability,
    AllPossible,
    StructureAsInteractions,
    StructureCompareMany,
    StructureCompareTwo,
)
from rnavigate.data.pdb import PDB
from rnavigate.data.profile import (
    DanceMaP,
    DeltaProfile,
    Profile,
    RNPMaP,
    SHAPEMaP,
)
from rnavigate.data.annotation import Annotation, Motif, ORFs, domains

__all__ = [
    # from data
    "Sequence",
    "Data",
    # from secondary_structure
    "SecondaryStructure",
    "StructureCoordinates",
    "SequenceCircle",
    # from alignment
    "set_alignment",
    "set_multiple_sequence_alignment",
    "lookup_alignment",
    "SequenceAlignment",
    "AlignmentChain",
    "StructureAlignment",
    # from colors
    "ScalarMappable",
    # from interactions
    "Interactions",
    "SHAPEJuMP",
    "RINGMaP",
    "PAIRMaP",
    "PairingProbability",
    "AllPossible",
    "StructureAsInteractions",
    "StructureCompareMany",
    "StructureCompareTwo",
    # from PDB
    "PDB",
    # from profile
    "Profile",
    "SHAPEMaP",
    "DanceMaP",
    "RNPMaP",
    "DeltaProfile",
    # from annotations
    "Annotation",
    "Motif",
    "ORFs",
    "domains",
]



"""
Unit and regression test for the fpocketR package.
"""

import sys
import subprocess
from pathlib import Path
import os
import pytest
import pandas as pd
import atexit


# --- Cleanup orphaned fpocket output files after tests ---
@pytest.fixture(scope="session", autouse=True)
def cleanup_fpocket_files():
    files_to_remove = [
        Path(__file__).parent.parent.parent / "2l1v_clean.pdb",
        Path(__file__).parent.parent.parent / "8f4o_clean.pdb"
    ]
    def remove_files():
        for f in files_to_remove:
            if f.exists():
                f.unlink()
    atexit.register(remove_files)


# --- Basic Import Test ---
def test_fpocketR_imported():
    """Sample test, will always pass so long as import statement worked."""
    assert "fpocketR" in sys.modules



# --- Helper Functions ---
def get_multistate_subdirs(multistate_dir):
    """Return sorted list of 2l1v_clean_stateX_out subdirectories."""
    return sorted([d for d in multistate_dir.iterdir() if d.is_dir() and d.name.startswith("2l1v_clean_state")])

def get_expected_state_dirnames():
    return [f"2l1v_clean_state{i}_out" for i in range(1, 4)]

# --- CSV Comparison Helper ---
def tolerant_csv_compare(file1, file2, atol=10):
    """Compare two CSVs using pandas, allowing small differences in float columns."""
    df1 = pd.read_csv(file1)
    df2 = pd.read_csv(file2)
    if df1.shape != df2.shape or list(df1.columns) != list(df2.columns):
        print(f"[DEBUG] CSV shape/columns mismatch: {file1} vs {file2}")
        return False
    # If 'Volume' column exists, allow higher tolerance for it
    if 'Volume' in df1.columns:
        try:
            # Compare all columns except 'Volume'
            cols = [c for c in df1.columns if c != 'Volume']
            pd.testing.assert_frame_equal(df1[cols], df2[cols], atol=atol, rtol=0, check_dtype=False, check_exact=False)
            # Compare 'Volume' with even higher tolerance
            pd.testing.assert_series_equal(df1['Volume'], df2['Volume'], atol=atol * 3, rtol=0, check_dtype=False, check_exact=False)
            return True
        except AssertionError as e:
            print(f"[DEBUG] CSV difference in {file1} vs {file2}:")
            print(e)
            return False
    else:
        try:
            pd.testing.assert_frame_equal(df1, df2, atol=atol, rtol=0, check_dtype=False, check_exact=False)
            return True
        except AssertionError as e:
            print(f"[DEBUG] CSV difference in {file1} vs {file2}:")
            print(e)
            return False


# --- Single-State Output Fixture (Alternate Command) ---
@pytest.fixture(scope="session")
def single_state_alt_output(tmp_path_factory):
    """Run fpocketR with alternate command and return output dir."""
    repo_root = Path(__file__).parent.parent.parent.resolve()
    data_dir = repo_root / "fpocketR" / "data"
    pdb_file = data_dir / "8f4o.pdb"
    al_file = data_dir / "2gdi.pdb"
    tmp_path = tmp_path_factory.mktemp("single_state_alt")
    out_dir = tmp_path / "8f4o_clean_out"
    out_dir.mkdir(parents=True, exist_ok=True)
    rel_out_dir = os.path.relpath(out_dir, repo_root)
    cmd = [
        "python", "-m", "fpocketR",
        "-pdb", str(pdb_file),
        "-al", str(al_file),
        "-l", "no",
        "-dpi", "50",
        "-o", rel_out_dir
    ]
    result = subprocess.run(cmd, check=True, cwd=str(repo_root), capture_output=True, text=True)
    return out_dir


# --- Alternate Single-State Output Tests ---
def test_single_state_alt_pdb_match_reference(single_state_alt_output):
    """Compare generated 8f4o_out_real_sphere.pdb to reference file."""
    out_dir = single_state_alt_output
    repo_root = Path(__file__).parent.parent.parent.resolve()
    data_dir = repo_root / "fpocketR" / "data"
    generated_pdbs = list(out_dir.rglob("8f4o_out_real_sphere.pdb"))
    assert generated_pdbs, f"No generated 8f4o_out_real_sphere.pdb file found in {out_dir}"
    generated_pdb = generated_pdbs[0]
    reference_pdb = data_dir / "TPP_apo_holo" / "8f4o_clean_out" / "8f4o_out_real_sphere.pdb"
    assert reference_pdb.exists(), f"Reference file not found: {reference_pdb}"
    with open(generated_pdb, "r") as f1, open(reference_pdb, "r") as f2:
        gen_lines = [line.strip() for line in f1 if line.strip()]
        ref_lines = [line.strip() for line in f2 if line.strip()]
    assert gen_lines == ref_lines, f"Generated PDB does not match reference PDB in {out_dir}"


def test_single_state_alt_3d_files_exist(single_state_alt_output):
    """Test that 3D files are generated for alternate single-state output."""
    out_dir = single_state_alt_output
    files = list(out_dir.rglob("*"))
    files = [f for f in files if f.is_file()]
    assert any(f.name.endswith("_3D_50.png") for f in files), f"No _3D_50.png in {out_dir}"
    assert any(f.name.endswith("_real_sphere.pdb") for f in files), f"No _real_sphere.pdb in {out_dir}"
    assert any(f.name.endswith("_real_sphere.pse") for f in files), f"No _real_sphere.pse in {out_dir}"


def test_single_state_alt_csv_match_reference(single_state_alt_output):
    """Compare generated *_pocket_characteristics.csv to reference file for alt single-state output."""
    out_dir = single_state_alt_output
    repo_root = Path(__file__).parent.parent.parent.resolve()
    data_dir = repo_root / "fpocketR" / "data"
    generated_csvs = list(out_dir.rglob("8f4o_out_pocket_characteristics.csv"))
    assert generated_csvs, f"No generated 8f4o_out_pocket_characteristics.csv file found in {out_dir}"
    generated_csv = generated_csvs[0]
    reference_csv = data_dir / "TPP_apo_holo" / "8f4o_clean_out" / "8f4o_out_pocket_characteristics.csv"
    assert reference_csv.exists(), f"Reference file not found: {reference_csv}"
    assert tolerant_csv_compare(generated_csv, reference_csv), f"Generated CSV does not match reference CSV in {out_dir}"


# --- Multistate Output Fixture ---

import pytest

@pytest.fixture(scope="session")
def multistate_output(tmp_path_factory):
    """Run the multistate fpocketR command ONCE and return output dir. Prints diagnostics and skips tests if fails."""
    tmp_path = tmp_path_factory.mktemp("multistate")
    repo_root = Path(__file__).parent.parent.parent.resolve()
    data_dir = repo_root / "fpocketR" / "data"
    pdb_file = (data_dir / "2l1v.pdb").resolve()
    ss_file = (data_dir / "2l1v.nsd").resolve()
    out_dir = tmp_path / "2l1v_multistate"
    out_dir.mkdir(parents=True, exist_ok=True)
    rel_out_dir = os.path.relpath(out_dir, repo_root)
    cmd = [
        "python", "-m", "fpocketR",
        "-pdb", str(pdb_file),
        "-ss", str(ss_file),
        "-s", "0", "-dpi", "50", "-o", rel_out_dir
    ]
    try:
        result = subprocess.run(cmd, check=True, cwd=str(repo_root), capture_output=True, text=True)
        return out_dir
    except subprocess.CalledProcessError as e:
        import sys
        sys.stderr.write("\n[ERROR] Multistate fpocketR subprocess failed!\n")
        sys.stderr.write(f"Command: {e.cmd}\n")
        sys.stderr.write(f"Return code: {e.returncode}\n")
        sys.stderr.write(f"Stdout:\n{e.output}\n")
        sys.stderr.write(f"Stderr:\n{e.stderr}\n")
        pytest.fail(f"Multistate fpocketR run failed: {e}")


# --- Multistate Output Tests ---
MULTISTATE_DIR = Path(__file__).parent.parent.parent / "fpocketR" / "data" / "preQ1_multistate"

@pytest.mark.parametrize("state_idx", range(1, 4))
def test_multistate_3d_files_exist(multistate_output, state_idx):
    """Test that 3D files exist in each state output directory."""
    out_dir = multistate_output
    gen_dir = out_dir / f"2l1v_clean_state{state_idx}_out"
    assert gen_dir.exists(), f"Missing generated subdir: {gen_dir}"
    files = [f for f in gen_dir.iterdir() if f.is_file()]
    assert any(f.name.endswith("_3D_50.png") for f in files), f"No _3D_50.png in {gen_dir}"
    assert any(f.name.endswith("_real_sphere.pdb") for f in files), f"No _real_sphere.pdb in {gen_dir}"
    assert any(f.name.endswith("_real_sphere.pse") for f in files), f"No _real_sphere.pse in {gen_dir}"


@pytest.mark.parametrize("state_idx", range(1, 4))
def test_multistate_2d_files_exist(multistate_output, state_idx):
    out_dir = multistate_output
    gen_dir = out_dir / f"2l1v_clean_state{state_idx}_out"
    assert gen_dir.exists(), f"Missing generated subdir: {gen_dir}"
    files = [f for f in gen_dir.iterdir() if f.is_file()]
    assert any(f.name.endswith("_2D.png") for f in files), f"No _2D.png in {gen_dir}"
    assert any(f.name.endswith("_2D.svg") for f in files), f"No _2D.svg in {gen_dir}"


@pytest.mark.parametrize("state_idx", range(1, 4))
def test_multistate_1d_csv_match_reference(multistate_output, state_idx):
    out_dir = multistate_output
    gen_dir = out_dir / f"2l1v_clean_state{state_idx}_out"
    ref_dir = MULTISTATE_DIR / f"2l1v_clean_state{state_idx}_out"
    gen_csvs = list(gen_dir.glob("*_pocket_characteristics.csv"))
    # Reference CSVs are named 2l1v_state{state_idx}_out_pocket_characteristics.csv in new structure
    ref_csvs = list(ref_dir.glob(f"2l1v_state{state_idx}_out_pocket_characteristics.csv"))
    assert gen_csvs, f"No generated *_pocket_characteristics.csv in {gen_dir}"
    assert ref_csvs, f"No reference 2l1v_state{state_idx}_out_pocket_characteristics.csv in {ref_dir}"
    gen_csv = gen_csvs[0]
    ref_csv = ref_csvs[0]
    assert tolerant_csv_compare(gen_csv, ref_csv), f"CSV mismatch in {gen_dir}"


def test_multistate_summary_files_exist(multistate_output):
    """Test that summary files exist in the output directory."""
    out_dir = multistate_output
    gen_files = [f for f in out_dir.iterdir() if f.is_file()]
    # Accept any file matching *_all_states_out_real_sphere.pse
    assert any(f.name.endswith("_all_states_out_real_sphere.pse") for f in gen_files), "Missing *_all_states_out_real_sphere.pse"
    assert any(f.name.endswith("_pocket_density.png") for f in gen_files), "Missing _pocket_density.png"
    assert any(f.name.endswith("all_states_pocket_characteristics.csv") for f in gen_files), "Missing all_states_pocket_characteristics.csv"


"""LogCompare compares reactivity profiles for significant differences.

This analysis requires replicates.
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize_scalar
from rnavigate.plots import Plot


class LogCompare:
    """Compares 2 experimental samples, given replicates of each sample.
    First, the log10(modified/untreated) rate is calculated. These values a
    scaled to minimize the median of the absolute value of the difference
    between samples. The standard error in these values is computed for each
    replicate. Z-scores between samples are calculated. The results are plotted
    in two panels: (1) the scaled log10(modified/untreated) rate for each
    sample with error bars, and (2) the difference between samples, colored by
    z-score.

    Methods:
        __init__: computes log10(modified/untreated) rates, rescales the data,
            then calls make_plot()
        get_profile_sequence: gets log10(m/u) rate and sequence from sample
        calc_scale_factor: gets scale factor given two profiles
        rescale: rescales a profile to minimize difference to another profile
        load_replicates: calculates average and standard error of replicates
        make_plots: displays the two panels described above.

    Attributes:
        data (str): a key of sample.data to retrieve per-nucleotide data
        groups (dict): a dictionary containing the following key-value pairs:
            1: a dictionary containing these key-value pairs:
                self.data: averaged scaled log10(m/u) across replicates
                "stderr": the standard errors across replicates
                "stacked": 2d array containing each scaled log10(m/u) array
                "seq": the sequence string
            2: same as 1 above, for the second sample
    """

    def __init__(self, samples1, samples2, name1, name2, data="profile", region="all"):
        """Takes replicates of two samples for comparison. Replicates are
        required. Calculates the log division profile
        (log10(modified/untreated)) and minimizes the median of the absolute
        difference between these. Finally, creates a plot.

        Args:
            samples1 (list of Sample objects): Replicates of the first sample
            samples2 (list of Sample objects): Replicates of the second sample
            name1 (string): name of first sample
            name2 (string): name of second sample
            data (str, optional): Datatype to compare. Defaults to "profile".
        """
        if region == "all":
            self.region = [0, samples1[0].data["profile"].length]
        else:
            self.region = region
        self.data = data
        self.groups = {}
        self.load_replicates(*samples1, group=1)
        self.load_replicates(*samples2, group=2)
        sequences_match = self.groups[1]["seq"] == self.groups[2]["seq"]
        assert sequences_match, "Sample sequences do not match."
        self.groups[2][self.data] = self.rescale(
            self.groups[2][self.data], self.groups[1][self.data]
        )
        self.groups[1]["name"] = name1
        self.groups[2]["name"] = name2
        self.make_plots()

    def get_profile_sequence(self, sample):
        """retrieves log10(Modified_rate/Untreated_rate) and sequence from
        sample.data[self.data].

        Args:
            sample (rnavigate.Sample): an rnavigate sample

        Returns:
            np.array, np.array: log10 profile and sequence
        """
        df = sample.data[self.data].data
        plus = df.Modified_rate.values.copy()
        minus = df.Untreated_rate.values.copy()
        profile = plus / minus
        profile = np.log(profile)
        profile[minus > 0.05] = np.nan
        return profile, sample.data[self.data].sequence

    def calc_scale_factor(self, profile, target_profile):
        """Finds the scale factor (x) that minimizes the median value of
        |profile + x - target_profile|.

        Args:
            profile (np.array): log10 profile
            target_profile (np.array): 2nd log10 profile
        """

        def f(offset):
            return np.nanmedian(np.abs(profile + offset - target_profile))

        result = minimize_scalar(f, bounds=[-20, 20])
        offset = result.x
        return offset

    def rescale(self, profile, target_profile):
        """scales profile to minimize difference to target_profile using
        calc_scale_factor

        Args:
            profile (np.array): log10 profile to scale
            target_profile (np.array): 2nd log10 profile

        Returns:
            np.array: scaled profile
        """
        offset = self.calc_scale_factor(profile, target_profile)
        return profile + offset

    def load_replicates(self, *samples, group):
        """calculates average and standard error for a group of replicates,
        stores these values, along with sequence and a np.array containing
        all profiles in self.groups[group].

        Args:
            *samples (list of rnavigate.Sample): replicates to load
            group (int): self.groups key to access replicate data
        """
        profile1, seq1 = self.get_profile_sequence(samples[0])
        primermask = np.array([c.islower() for c in seq1])
        profile1[primermask] = np.nan

        profiles = [profile1]
        for sample in samples[1:]:
            profile, seq = self.get_profile_sequence(sample)
            assert seq == seq1, "Replicate sequences do not match."
            profile[primermask] = np.nan
            profile = self.rescale(profile, profile1)
            profiles.append(profile)
        stacked = np.vstack(profiles)
        avgprofile = np.nanmean(stacked, ax=0)
        stderr = np.std(stacked, ax=0)
        self.groups[group] = {
            self.data: avgprofile,
            "stderr": stderr,
            "stacked": stacked,
            "seq": seq,
        }

    def make_plots(self):
        """Visualize this analysis."""
        # get the replicate data
        prof1, stderr1, stack1, seq, name1 = self.groups[1].values()
        prof2, stderr2, stack2, seq, name2 = self.groups[2].values()

        x = np.array(list(range(1, len(prof1) + 1)))

        # create a two row, one column figure, scaled for RNA length
        _, axes = plt.subplots(2, 1, figsize=(0.1 * len(prof1), 14), sharex=True)
        # first axes contains raw log10 profiles with error bars
        ax = axes[0]
        ax.step(x, prof1, label=name1, color="C0")
        ax.step(x, prof2, label=name2, color="C1")
        ax.fill_between(
            x,
            prof1 - stderr1,
            prof1 + stderr1,
            step="pre",
            color="C0",
            alpha=0.25,
            lw=0,
        )
        ax.fill_between(
            x,
            prof2 - stderr2,
            prof2 + stderr2,
            step="pre",
            color="C1",
            alpha=0.25,
            lw=0,
        )
        ax.legend(loc="upper right")
        ax.axhline(0, color="black", lw=1, zorder=0)
        ax.set_ylabel("ln(Mod/BG)")
        Plot.add_sequence(self, ax=ax, sequence=seq)

        # calculate mean difference and standard error for z-scores
        stack1 = stack1 + 10
        rescale_dms = self.rescale(stack1, stack2)
        stack1 = rescale_dms
        diff = stack2 - stack1
        meandiff = np.nanmean(diff, ax=0)
        std_err = np.std(diff, ax=0)
        # compute z-scores
        z_scores = meandiff / std_err
        # normalize to 0-1 (values > 5 = 1, values < -5 = 0)
        z_scores = (z_scores + 5) / 10
        z_scores[z_scores > 1] = 1
        z_scores[z_scores < 0] = 0
        # get colormapping for differences
        # blue, white, red gradient from -5 to +5 stderrs
        colormap = plt.get_cmap("bwr")

        # create 2nd plot of differences colored by z-score
        ax = axes[1]
        ax.bar(x - 0.5, meandiff, color=colormap(z_scores), lw=0)
        ax.axhline(0, color="black", lw=1)
        ax.set_xlabel("nucleotide")
        ax.set_ylabel(f"{name2}-{name1}")
        ax.errorbar(x - 0.5, meandiff, yerr=std_err, fmt="none", color="black", lw=1)
        Plot.add_sequence(self, ax=ax, sequence=seq)

        # create a color bar scale for z-score differences
        axin1 = ax.inset_axes([0.8, 0.1, 0.15, 0.15])
        cmap = plt.get_cmap("bwr")
        Plot.view_colormap(
            ax=axin1,
            ticks=[0, 5, 10],
            values=[-5, 0, 5],
            title="Z-score",
            cmap=list(cmap(np.arange(cmap.N))),
        )
        plt.tight_layout()
